# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




from __future__ import division
import unittest




class TestParallel(unittest.TestCase):
    def test_convergence_advec(self):
        """Test whether 2/3D advection actually converges"""

        import numpy
        import numpy.linalg as la
        from hedge.mesh import make_ball_mesh, make_box_mesh, make_rect_mesh
        from hedge.discretization import Discretization, pair_with_boundary, norm
        from hedge.element import TetrahedralElement, TriangularElement
        from hedge.timestep import RK4TimeStepper
        from hedge.tools import EOCRecorder
        from math import sin, pi, sqrt
        from hedge.operators import StrongAdvectionOperator
        from hedge.data import TimeDependentGivenFunction
        from hedge.parallel import guess_parallelization_context
        from hedge.visualization import SiloVisualizer

        pcon = guess_parallelization_context()

        debug_output = True

        # note: x component must remain zero because x-periodicity is used
        v = numpy.array([0.0,0.9,0.3])

        def f(x):
            return sin(x)

        def u_analytic(x, t):
            return f((numpy.dot(-v[:dims],x)/la.norm(v[:dims])+t*la.norm(v[:dims])))

        def boundary_tagger(vertices, el, face_nr):
            face_normal = el.face_normals[face_nr]
            if numpy.dot(face_normal, v[:len(face_normal)]) < 0:
                return ["inflow"]
            else:
                return ["outflow"]

        for i_mesh, mesh in enumerate([
            # 2D semiperiodic
            make_rect_mesh(b=(2*pi,3), max_area=0.4,
                periodicity=(True, False),
                subdivisions=(5,10),
                boundary_tagger=boundary_tagger, 
                ),
            # 3D x-periodic
            make_box_mesh(dimensions=(2*pi, 2, 2), max_volume=0.4,
                periodicity=(True, False, False),
                boundary_tagger=boundary_tagger, 
                ),
            # non-periodic
            make_ball_mesh(r=pi, 
                boundary_tagger=boundary_tagger, max_volume=0.7),
            ]):
            for flux_type in StrongAdvectionOperator.flux_types:
                for random_partition in [True, False]:
                    eoc_rec = EOCRecorder()

                    if random_partition:
                        # Distribute elements randomly across nodes.
                        # This is bad, efficiency-wise, but it puts stress
                        # on the parallel implementation, which is desired here.
                        # Another main point of this is to force the code to split
                        # a periodic face pair across nodes.
                        from random import choice
                        partition = [choice(pcon.ranks) for el in mesh.elements]
                    else:
                        partition = None

                    for order in [1,2,3,4]:
                        if pcon.is_head_rank:
                            mesh_data = pcon.distribute_mesh(mesh, partition)
                        else:
                            mesh_data = pcon.receive_mesh()

                        dims = mesh.points.shape[1]
                        if dims == 2:
                            el_class = TriangularElement
                        else:
                            el_class = TetrahedralElement

                        discr = pcon.make_discretization(mesh_data, 
                                el_class(order), debug=True)

                        op = StrongAdvectionOperator(discr, v[:dims], 
                                inflow_u=TimeDependentGivenFunction(u_analytic),
                                flux_type=flux_type)
                        vis = SiloVisualizer(discr, pcon)

                        u = discr.interpolate_volume_function(lambda x: u_analytic(x, 0))
                        ic = u.copy()

                        dt = discr.dt_factor(op.max_eigenvalue())
                        nsteps = int(1/dt)
                        if debug_output and pcon.is_head_rank:
                            print "#steps=%d #elements=%d" % (nsteps, len(mesh.elements))

                        test_name = "test-%s-o%d-m%d-r%s" % (
                                flux_type, order, i_mesh, random_partition)

                        stepper = RK4TimeStepper()
                        for step in range(nsteps):
                            u = stepper(u, step*dt, dt, op.rhs)

                        u_true = discr.interpolate_volume_function(
                                lambda x: u_analytic(x, nsteps*dt))
                        error = u-u_true
                        my_l2_error = norm(discr, error)

                        if debug_output:
                            visf = vis.make_file(test_name+"-final")
                            vis.add_data(visf, [
                                ("u", u),
                                ("u_true", u_true),
                                ("ic", ic)])
                            visf.close()

                        from operator import add
                        from boost.mpi import all_reduce
                        l2_error = all_reduce(pcon.communicator, my_l2_error, add)

                        eoc_rec.add_data_point(order, l2_error)

                    if debug_output and pcon.is_head_rank:
                        print "%s\n%s\n" % (flux_type.upper(), "-" * len(flux_type))
                        print eoc_rec.pretty_print(abscissa_label="Poly. Order", 
                                error_label="L2 Error")

                    self.assert_(eoc_rec.estimate_order_of_convergence()[0,1] > 3)
                    self.assert_(eoc_rec.estimate_order_of_convergence(2)[-1,1] > 7)





if __name__ == '__main__':
    unittest.main()



"""Logging utilities."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""



from pytools.log import LogQuantity, MultiLogQuantity
import numpy
import numpy.linalg as la




def axis_name(axis):
    if axis == 0: return "x"
    elif axis == 1: return "y"
    elif axis == 2: return "z"
    else: raise RuntimeError, "invalid axis index"




class Integral(LogQuantity):
    """Log the volume integral of a variable in a scope."""

    def __init__(self, getter, discr, name=None, 
            unit="1", description=None):
        """Construct the integral logger.

        @arg getter: a callable that returns the value of which to 
          take the integral.
        @arg discr: a L{Discretization} to which the variable belongs.
        @arg name: the name reported to the C{LogManager}.
        @arg unit: the unit of measure for the log quantity.
        @arg description: A description fed to the C{LogManager}.
        """
        self.getter = getter

        if name is None:
            try:
                name = "int_%s" % self.getter.name()
            except AttributeError:
                raise ValueError, "must specify a name"

        LogQuantity.__init__(self, name, unit, description)

        self.discr = discr

    @property
    def default_aggregator(self): 
        return sum

    def __call__(self):
        var = self.getter()

        from hedge.discretization import integral
        from hedge.tools import log_shape

        if len(log_shape(var)) == 1:
            return sum(
                    integral(self.discr, numpy.abs(v))
                    for v in var)
        else:
            return integral(self.discr, var)




class LpNorm(LogQuantity):
    """Log the Lp norm of a variable in a scope."""

    def __init__(self, getter, discr, p=2, name=None, 
            unit="1", description=None):
        """Construct the Lp norm logger.

        @arg getter: a callable that returns the value of which to 
          take the norm.
        @arg discr: a L{Discretization} to which the variable belongs.
        @arg p: the power of the norm.
        @arg name: the name reported to the C{LogManager}.
        @arg unit: the unit of measure for the log quantity.
        @arg description: A description fed to the C{LogManager}.
        """
        self.getter = getter
        self.discr = discr
        self.p = p

        if name is None:
            try:
                name = "l%d_%s" % (int(p), self.getter.name())
            except AttributeError:
                raise ValueError, "must specify a name"

        LogQuantity.__init__(self, name, unit, description)

    @property
    def default_aggregator(self): 
        from pytools import norm_inf, Norm

        if self.p == numpy.Inf:
            return norm_inf
        else:
            return Norm(self.p)

    def __call__(self):
        var = self.getter()

        from hedge.discretization import norm
        from hedge.tools import log_shape

        return norm(self.discr, var, self.p)




# electromagnetic quantities --------------------------------------------------
class EMFieldGetter(object):
    """Makes E and H field accessible as self.e and self.h from a variable lookup.
    To be used with the EM log quantities in this module."""
    def __init__(self, maxwell_op, scope, varname):
        self.maxwell_op = maxwell_op
        self.scope = scope
        self.varname = varname

    @property
    def e(self):
        fields = self.scope[self.varname]
        e, h = self.maxwell_op.split_eh(fields)
        return e

    @property
    def h(self):
        fields = self.scope[self.varname]
        e, h = self.maxwell_op.split_eh(fields)
        return h




class EMFieldEnergy(LogQuantity):
    def __init__(self, fields, name="W_field"):
        LogQuantity.__init__(self, name, "J", "Field Energy")
        self.fields = fields

    @property
    def default_aggregator(self): 
        from pytools import norm_2
        return norm_2

    def __call__(self):
        max_op = self.fields.maxwell_op

        e = self.fields.e
        h = self.fields.h
        d = max_op.epsilon * e
        b = max_op.mu * h

        from hedge.tools import ptwise_dot
        energy_density = 1/2*(
                ptwise_dot(e, d) 
                + ptwise_dot(h, b))

        from hedge.discretization import integral
        return integral(max_op.discr, energy_density)




class EMFieldMomentum(MultiLogQuantity):
    def __init__(self, fields, c0, names=None):
        h_cross = fields.maxwell_op.h_cross
        if names is None:
            names = ["p%s_field" % axis_name(i) 
                    for i, included in enumerate(h_cross.result_subset)
                    if included]

        vdim = len(names)

        MultiLogQuantity.__init__(self, names, 
            units=["N*s"] * vdim, 
            descriptions=["Field Momentum"] * vdim)

        self.fields = fields
        self.c0 = c0

    def __call__(self):
        max_op = self.fields.maxwell_op

        e = self.fields.e
        h = self.fields.h

        poynting_s = max_op.h_cross(e, h)

        momentum_density = poynting_s/self.c0**2

        from hedge.discretization import integral
        return integral(max_op.discr, momentum_density)




class EMFieldDivergenceD(LogQuantity):
    def __init__(self, maxwell_op, fields, name="divD"):
        LogQuantity.__init__(self, name, "C", "Integral over div D")

        self.fields = fields
        self.discr = self.fields.maxwell_op.discr

        from hedge.operators import DivergenceOperator
        self.div_op = DivergenceOperator(self.discr,
                maxwell_op.get_eh_subset()[:3])

    def __call__(self):
        max_op = self.fields.maxwell_op
        d = max_op.epsilon * self.fields.e
        div_d = self.div_op(d)
        
        from hedge.discretization import integral
        return integral(self.discr, div_d)




class EMFieldDivergenceB(MultiLogQuantity):
    def __init__(self, maxwell_op, fields, names=None):
        self.fields = fields
        self.discr = self.fields.maxwell_op.discr

        from hedge.operators import DivergenceOperator
        self.div_op = DivergenceOperator(self.discr,
                maxwell_op.get_eh_subset()[3:])

        if names is None:
            names = ["divB", "err_divB_l1"]

        MultiLogQuantity.__init__(self, 
                names=names,
                units=["T/m", "T/m"], 
                descriptions=["Integral over div B", "Integral over |div B|"])

    def __call__(self):
        max_op = self.fields.maxwell_op
        b = max_op.mu * self.fields.h
        div_b = self.div_op(b)
        
        from hedge.discretization import integral
        return [integral(self.discr, div_b), integral(self.discr, numpy.absolute(div_b))]




def add_em_quantities(mgr, maxwell_op, fields):
    mgr.add_quantity(EMFieldEnergy(fields))
    mgr.add_quantity(EMFieldMomentum(fields, maxwell_op.c))
    mgr.add_quantity(EMFieldDivergenceD(maxwell_op, fields))
    mgr.add_quantity(EMFieldDivergenceB(maxwell_op, fields))



"""1D Newton interpolation helper module."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import pymbolic




def newton_interpolation_coefficients(x, y):
    assert len(x) == len(y)
    n = len(y)
    divided_differences = [y]
    last = y

    for step in range(1, n):
        next = [(last[i+1]-last[i])/(x[i+step]-x[i])
                for i in range(n-step)]
        divided_differences.append(next)
        last = next

    return [dd_col[-1] for dd_col in divided_differences]




def newton_interpolation_polynomial(x, y):
    coeff = newton_interpolation_coefficients(x, y)

    var_x = pymbolic.var("x")
    linear_factors = [
            pymbolic.Polynomial(var_x, ((0, pt), (1, 1)))
            for pt in x]
    pyramid_linear_factors = [1]
    for l in linear_factors:
        pyramid_linear_factors.append(
                pyramid_linear_factors[-1]*l)

    return pymbolic.linear_combination(coeff, pyramid_linear_factors)




def newton_interpolation_function(x, y):
    return pymbolic.compile(newton_interpolation_polynomial(x, y), ["x"])



"""Global function space discretization."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""




import numpy
import numpy.linalg as la
import pyublas
import hedge.tools
import hedge.mesh
import hedge._internal




class _FaceGroup(hedge._internal.FaceGroup):
    def __init__(self, double_sided):
        hedge._internal.FaceGroup.__init__(self, double_sided)
        self.face_index_lists = []
        self.face_index_list_register = {}

    def register_face_indices(self, identifier, generator):

        try:
            return self.face_index_list_register[identifier]
        except KeyError:
            new_idx = len(self.face_index_lists)
            fil = generator()
            self.face_index_lists.append(fil)
            self.face_index_list_register[identifier] = new_idx
            return new_idx

    def commit_face_index_lists(self):
        from hedge._internal import IntVector

        for fil in self.face_index_lists:
            intvec = IntVector(fil)

            # allow prefetching a few entries past the end
            for i in range(4):
                intvec.append(intvec[-1])

            self.index_lists.append(intvec)

        del self.face_index_lists
        del self.face_index_list_register





class _ElementGroup(object):
    """Once fully filled, this structure has the following data members:

    @ivar members: a list of hedge.mesh.Element instances in this group.
    @ivar local_discretization: an instance of hedge.element.Element.
    @ivar ranges: a list of C{(start, end)} tuples indicating the DOF numbers for
      each element. Note: This is actually a C++ ElementRanges object.
    @ivar mass_matrix: The element-local mass matrix M{M}.
    @ivar inverse_mass_matrix: the element-local inverese mass matrix M{M^{-1}}.
    @ivar differentiation_matrices: local differentiation matrices M{D_r, D_s, D_t}, 
      i.e.  differentiation by M{r, s, t, ....}.
    @ivar stiffness_matrices: the element-local stiffness matrices M{M*D_r, M*D_s,...}.
    @ivar jacobians: list of jacobians over all elements
    @ivar inverse_jacobians: inverses of L{jacobians}.
    @ivar diff_coefficients: a M{(d,d)}-matrix of coefficient vectors to turn
      M{(r,s,t)}-differentiation into M{(x,y,z)}.
    """
    pass




class _Boundary(object):
    def __init__(self, nodes, ranges, index_map, face_groups_and_ldis,
            el_face_to_face_group_and_flux_face_index={}):
        self.nodes = nodes
        self.ranges = ranges
        self.index_map = index_map
        self.face_groups_and_ldis = face_groups_and_ldis
        self.el_face_to_face_group_and_flux_face_index = \
                el_face_to_face_group_and_flux_face_index




class Discretization(object):
    """The global approximation space.

    Instances of this class tie together a local discretization (i.e. polynomials
    on an elemnent) into a function space on a mesh. They provide creation
    functions such as interpolating given functions, differential operators and
    flux lifting operators.
    """

    def __init__(self, mesh, local_discretization, debug=False):
        self.mesh = mesh

        self.dimensions = local_discretization.dimensions
        self.debug = debug

        self._build_element_groups_and_nodes(local_discretization)
        self._calculate_local_matrices()
        self._build_interior_face_groups()
        self.boundaries = {}

        # instrumentation -----------------------------------------------------
        from pytools.log import IntervalTimer, EventCounter

        self.inner_flux_counter = EventCounter("n_inner_flux", 
                "Number of inner flux computations")
        self.inner_flux_timer = IntervalTimer("t_inner_flux", 
                "Time spent computing inner fluxes")
        self.bdry_flux_counter = EventCounter("n_bdry_flux", 
                "Number of boundary flux computations")
        self.bdry_flux_timer = IntervalTimer("t_bdry_flux", 
                "Time spent computing boundary fluxes")

        self.mass_op_counter = EventCounter("n_mass_op", 
                "Number of mass operator applications")
        self.mass_op_timer = IntervalTimer("t_mass_op", 
                "Time spent applying mass operators")
        self.diff_op_counter = EventCounter("n_diff_op",
                "Number of differentiation operator applications")
        self.diff_op_timer = IntervalTimer("t_diff_op",
                "Time spent applying applying differentiation operators")

        self.interpolant_counter = EventCounter("n_interp", 
                "Number of interpolant evaluations")
        self.interpolant_timer = IntervalTimer("t_interp", 
                "Time spent evaluating interpolants")

        self.instrumented = False

    # instrumentation ---------------------------------------------------------
    def add_instrumentation(self, mgr):
        mgr.add_quantity(self.inner_flux_counter)
        mgr.add_quantity(self.inner_flux_timer)
        mgr.add_quantity(self.bdry_flux_counter)
        mgr.add_quantity(self.bdry_flux_timer)
        mgr.add_quantity(self.mass_op_counter)
        mgr.add_quantity(self.mass_op_timer)
        mgr.add_quantity(self.diff_op_counter)
        mgr.add_quantity(self.diff_op_timer)
        mgr.add_quantity(self.interpolant_counter)
        mgr.add_quantity(self.interpolant_timer)

        from pytools.log import time_and_count_function
        self.interpolate_volume_function = \
                time_and_count_function(
                        self.interpolate_volume_function,
                        self.interpolant_timer,
                        self.interpolant_counter)

        self.interpolate_boundary_function = \
                time_and_count_function(
                        self.interpolate_boundary_function,
                        self.interpolant_timer,
                        self.interpolant_counter)

        self.perform_inner_flux = \
                time_and_count_function(
                        self.perform_inner_flux,
                        self.inner_flux_timer,
                        self.inner_flux_counter)

        self._perform_boundary_flux = \
                time_and_count_function(
                        self._perform_boundary_flux,
                        self.bdry_flux_timer,
                        self.bdry_flux_counter)

        self.instrumented = True

    # initialization ----------------------------------------------------------
    def _build_element_groups_and_nodes(self, local_discretization):
        from hedge._internal import UniformElementRanges

        eg = _ElementGroup()
        eg.members = self.mesh.elements
        eg.local_discretization = ldis = local_discretization
        eg.ranges = UniformElementRanges(
                0, 
                len(ldis.unit_nodes()), 
                len(self.mesh.elements))

        nodes_per_el = ldis.node_count()
        # mem layout:
        # [....element....][...element...]
        #  |    \
        #  [node.]
        #   | | |
        #   x y z

        self.nodes = numpy.empty(
                (len(self.mesh.elements)*nodes_per_el, self.dimensions),
                dtype=float, order="C")

        unit_nodes = numpy.empty( (nodes_per_el, self.dimensions),
                dtype=float, order="C")

        for i_node, node in enumerate(ldis.unit_nodes()):
            unit_nodes[i_node] = node

        from hedge._internal import map_element_nodes

        for el in self.mesh.elements:
            map_element_nodes(
                    self.nodes,
                    el.id*nodes_per_el*self.dimensions,
                    el.map,
                    unit_nodes,
                    self.dimensions)

        self.group_map = [(eg, i) for i in range(len(self.mesh.elements))]
        self.element_groups = [eg]

    def _calculate_local_matrices(self):
        for eg in self.element_groups:
            ldis = eg.local_discretization

            mmat = eg.mass_matrix = ldis.mass_matrix()
            immat = eg.inverse_mass_matrix = ldis.inverse_mass_matrix()
            dmats = eg.differentiation_matrices = \
                    ldis.differentiation_matrices()
            smats = eg.stiffness_matrices = numpy.array(
                    [numpy.dot(mmat, d) for d in dmats])
            smats = eg.stiffness_t_matrices = numpy.array(
                    [numpy.dot(d.T, mmat.T) for d in dmats])
            eg.minv_st = numpy.array(
                    [numpy.dot(numpy.dot(immat,d.T), mmat) for d in dmats])

            eg.jacobians = numpy.array([
                abs(el.map.jacobian) 
                for el in eg.members])
            eg.inverse_jacobians = numpy.array([
                abs(el.inverse_map.jacobian) 
                for el in eg.members])

            eg.diff_coefficients = [ # over global diff. coordinate
                    [ # local diff. coordinate
                        numpy.array([
                            el.inverse_map
                            .matrix[loc_coord, glob_coord]
                            for el in eg.members
                            ])
                        for loc_coord in range(ldis.dimensions)
                        ]
                    for glob_coord in range(ldis.dimensions)
                    ]

            eg.stiffness_coefficients = [ # over global diff. coordinate
                    [ # local diff. coordinate
                        numpy.array([
                            abs(el.map.jacobian)*el.inverse_map
                            .matrix[loc_coord, glob_coord]
                            for el in eg.members
                            ])
                        for loc_coord in range(ldis.dimensions)
                        ]
                    for glob_coord in range(ldis.dimensions)
                    ]

    def _make_flux_face(self, ldis, (el, fi)):
        from hedge.flux import FluxFace

        f = FluxFace()

        f.face_jacobian = el.face_jacobians[fi]
        f.element_id = el.id
        f.face_id = fi
        f.order = ldis.order
        f.normal = pyublas.why_not(el.face_normals[fi])

        # This crude approximation is shamelessly stolen from sledge.
        # There's an important caveat, however (which took me the better
        # part of a week to figure out):
        # h on both sides of an interface must be the same, otherwise
        # the penalty term will behave very oddly.
        f.h = abs(el.map.jacobian/f.face_jacobian)

        return f

    def _build_interior_face_groups(self):
        from hedge._internal import FaceGroup, FacePair
        from hedge.element import FaceVertexMismatch

        fg = _FaceGroup(double_sided=True)

        # find and match node indices along faces
        for i, (local_face, neigh_face) in enumerate(self.mesh.interfaces):
            e_l, fi_l = local_face
            e_n, fi_n = neigh_face

            (estart_l, eend_l), ldis_l = self.find_el_data(e_l.id)
            (estart_n, eend_n), ldis_n = self.find_el_data(e_n.id)

            vertices_l = e_l.faces[fi_l]
            vertices_n = e_n.faces[fi_n]

            findices_l = ldis_l.face_indices()[fi_l]
            findices_n = ldis_n.face_indices()[fi_n]

            try:
                findices_shuffle_op_n = \
                        ldis_l.get_face_index_shuffle_to_match(
                        vertices_l, vertices_n)

                if self.debug:
                    findices_shuffled_n = findices_shuffle_op_n(findices_n)

                    for i, j in zip(findices_l, findices_shuffled_n):
                        dist = self.nodes[estart_l+i]-self.nodes[estart_n+j]
                        assert la.norm(dist) < 1e-14

            except FaceVertexMismatch:
                # this happens if vertices_l is not a permutation of vertices_n.
                # periodicity is the only reason why that would be so.

                vertices_n, axis = self.mesh.periodic_opposite_faces[vertices_n]

                findices_shuffle_op_n = \
                        ldis_l.get_face_index_shuffle_to_match(vertices_l, vertices_n)

                if self.debug:
                    findices_shuffled_n = findices_shuffle_op_n(findices_n)

                    for i, j in zip(findices_l, findices_shuffled_n):
                        dist = self.nodes[estart_l+i]-self.nodes[estart_n+j]
                        dist[axis] = 0 
                        assert la.norm(dist) < 1e-14

            # create and fill the face pair
            fp = FacePair()

            fp.el_base_index = estart_l
            fp.opp_el_base_index = estart_n

            fp.face_index_list_number = fg.register_face_indices(
                    identifier=fi_l, 
                    generator=lambda: findices_l)
            fp.opp_face_index_list_number = fg.register_face_indices(
                    identifier=(fi_n, findices_shuffle_op_n),
                    generator=lambda : findices_shuffle_op_n(findices_n))

            fp.flux_face_index = len(fg.flux_faces)
            fp.opp_flux_face_index = len(fg.flux_faces)+1

            fg.face_pairs.append(fp)

            # create the flux faces
            flux_face_l = self._make_flux_face(ldis_l, local_face)
            flux_face_n = self._make_flux_face(ldis_n, neigh_face)

            # unify h across the faces
            flux_face_l.h = flux_face_n.h = max(flux_face_l.h, flux_face_n.h)

            fg.flux_faces.append(flux_face_l)
            fg.flux_faces.append(flux_face_n)

        fg.commit_face_index_lists()

        if len(fg.face_pairs):
            self.face_groups = [(fg, ldis_l.face_mass_matrix())]
        else:
            self.face_groups = []
        
    def _get_boundary(self, tag):
        """Get a _Boundary instance for a given `tag'.

        If there is no boundary tagged with `tag', an empty _Boundary instance
        is returned. Asking for a nonexistant boundary is not an error. 
        (Otherwise get_boundary would unnecessarily become non-local when run 
        in parallel.)
        """
        try:
            return self.boundaries[tag]
        except KeyError:
            pass

        from hedge._internal import IndexMap, FacePair

        nodes = []
        face_ranges = {}
        index_map = []
        face_group = _FaceGroup(double_sided=False)
        ldis = None # if this boundary is empty, we might as well have no ldis
        el_face_to_face_group_and_flux_face_index = {}

        for ef in self.mesh.tag_to_boundary.get(tag, []):
            el, face_nr = ef

            (el_start, el_end), ldis = self.find_el_data(el.id)
            face_indices = ldis.face_indices()[face_nr]

            f_start = len(nodes)
            nodes += [self.nodes[el_start+i] for i in face_indices]
            face_ranges[ef] = (f_start, len(nodes))
            index_map.extend(el_start+i for i in face_indices)

            # create the face pair
            fp = FacePair()
            fp.el_base_index = el_start
            fp.opp_el_base_index = f_start
            fp.face_index_list_number = face_group.register_face_indices(
                    identifier=face_nr,
                    generator=lambda: face_indices)
            fp.opp_face_index_list_number = face_group.register_face_indices(
                    identifier=(),
                    generator=lambda: tuple(xrange(len(face_indices))))
            fp.flux_face_index = len(face_group.flux_faces)
            fp.opp_flux_face_index = FacePair.INVALID_INDEX
            face_group.face_pairs.append(fp)

            # create the flux face
            face_group.flux_faces.append(self._make_flux_face(ldis, ef))
            
            # and make it possible to find it later
            el_face_to_face_group_and_flux_face_index[ef] = \
                    face_group, len(face_group.flux_faces)-1

        face_group.commit_face_index_lists()

        bdry = _Boundary(
                nodes=nodes,
                ranges=face_ranges,
                index_map=IndexMap(len(self.nodes), len(index_map), index_map),
                face_groups_and_ldis=[(face_group, ldis)],
                el_face_to_face_group_and_flux_face_index=
                el_face_to_face_group_and_flux_face_index)

        self.boundaries[tag] = bdry
        return bdry

    # vector construction -----------------------------------------------------
    def __len__(self):
        """Return the number of nodes in this discretization."""
        return len(self.nodes)

    def len_boundary(self, tag):
        return len(self._get_boundary(tag).nodes)

    def volume_empty(self, shape=()):
        return numpy.empty(shape+(len(self.nodes),), dtype=float)

    def volume_zeros(self, shape=()):
        return numpy.zeros(shape+(len(self.nodes),), dtype=float)

    def interpolate_volume_function(self, f):
        try:
            # are we interpolating many fields at once?
            shape = f.shape
        except AttributeError:
            # no, just one
            result = self.volume_zeros()
            for i, x in enumerate(self.nodes):
                result[i] = f(x)
            return result
        else:
            result = self.volume_zeros(shape)
            for point_nr, x in enumerate(self.nodes):
                result[(slice(None),)*len(shape) + (point_nr,)] = f(x)
            return result

    def boundary_zeros(self, tag=hedge.mesh.TAG_ALL, shape=()):
        return numpy.zeros(shape+(len(self._get_boundary(tag).nodes),),
                dtype=float)

    def interpolate_boundary_function(self, f, tag=hedge.mesh.TAG_ALL):
        try:
            # are we interpolating many fields at once?
            shape = f.shape
        except AttributeError:
            # no, just one
            result = self.boundary_zeros(tag=tag)
            for i, x in enumerate(self._get_boundary(tag).nodes):
                result[i] = f(x)
            return result
        else:
            result = self.boundary_zeros(tag=tag, shape=shape)
            for point_nr, x in enumerate(self._get_boundary(tag).nodes):
                result[(slice(None),)*len(shape) + (point_nr,)] = f(x)
            return result

    def boundary_normals(self, tag=hedge.mesh.TAG_ALL):
        result = self.boundary_zeros(tag=tag, shape=(self.dimensions,))
        for fg, ldis in self._get_boundary(tag).face_groups_and_ldis:
            for face_pair in fg.face_pairs:
                flux_face = fg.flux_faces[face_pair.flux_face_index]
                oeb = face_pair.opp_el_base_index
                opp_index_list = fg.index_lists[face_pair.opp_face_index_list_number]
                for i in opp_index_list:
                    result[:,oeb+i] = flux_face.normal

        return result

    # element data retrieval --------------------------------------------------
    def find_el_range(self, el_id):
        group, idx = self.group_map[el_id]
        return group.ranges[idx]

    def find_el_discretization(self, el_id):
        return self.group_map[el_id][0].local_discretization

    def find_el_data(self, el_id):
        group, idx = self.group_map[el_id]
        return group.ranges[idx], group.local_discretization

    # inner flux computation --------------------------------------------------
    def perform_inner_flux(self, int_flux, ext_flux, target):
        """Perform fluxes in the interior of the domain on the 
        given operator target.  This performs the contribution::

          M_{i,j} := \sum_{interior faces f} \int_f
            (  
               int_flux(f) * \phi_j
               + 
               ext_flux(f) * \phi_{opp(j)}
             )
             \phi_i
        
        on the given target. opp(j) denotes the dof with its node
        opposite from node j on the face opposite f.

        Thus the matrix product M*u, where u is the full volume field
        results in::

          v_i := M_{i,j} u_j
            = \sum_{interior faces f} \int_f
            (  
               int_flux(f) * u^-
               + 
               ext_flux(f) * u^+
             )
             \phi_i

        For more on operator targets, see src/cpp/op_target.hpp.

        Both local_flux and neighbor_flux must be instances of
        hedge.flux.Flux, i.e. compiled fluxes. Typically, you will
        not call this routine, it will be called for you by flux
        operators obtained by get_flux_operator().
        """
        from hedge._internal import perform_flux_on_one_target, ChainedFlux, NullTarget

        if isinstance(target, NullTarget):
            return

        ch_int = ChainedFlux(int_flux)
        ch_ext = ChainedFlux(ext_flux)

        target.begin(len(self.nodes), len(self.nodes))
        for fg, fmm in self.face_groups:
            perform_flux_on_one_target(fg, fmm, ch_int, ch_ext, target)
        target.finalize()

    # boundary flux computation -----------------------------------------------
    def _perform_boundary_flux(self, 
            int_flux, int_target, 
            ext_flux, ext_target, 
            bdry):
        from hedge._internal import perform_flux, ChainedFlux

        ch_int = ChainedFlux(int_flux)
        ch_ext = ChainedFlux(ext_flux)

        int_target.begin(len(self.nodes), len(self.nodes))
        ext_target.begin(len(self.nodes), len(bdry.nodes))
        if bdry.nodes:
            for fg, ldis in bdry.face_groups_and_ldis:
                perform_flux(fg, ldis.face_mass_matrix(), 
                        ch_int, int_target, 
                        ch_ext, ext_target)
        int_target.finalize()
        ext_target.finalize()

    # misc stuff --------------------------------------------------------------
    def dt_non_geometric_factor(self):
        distinct_ldis = set(eg.local_discretization for eg in self.element_groups)
        return min(ldis.dt_non_geometric_factor() 
                for ldis in distinct_ldis)

    def dt_geometric_factor(self):
        return min(min(eg.local_discretization.dt_geometric_factor(
            [self.mesh.points[i] for i in el.vertex_indices], el)
            for el in eg.members)
            for eg in self.element_groups)

    def dt_factor(self, max_system_ev):
        return 1/max_system_ev \
                * self.dt_non_geometric_factor() \
                * self.dt_geometric_factor()

    def volumize_boundary_field(self, bfield, tag=hedge.mesh.TAG_ALL):
        from hedge._internal import perform_inverse_index_map
        from hedge.tools import make_vector_target, log_shape

        ls = log_shape(bfield)
        result = self.volume_zeros(ls)
        bdry = self._get_boundary(tag)

        if ls != ():
            from pytools import indices_in_shape
            for i in indices_in_shape(ls):
                target = make_vector_target(bfield[i], result[i])
                target.begin(len(self.nodes), len(bdry.nodes))
                perform_inverse_index_map(bdry.index_map, target)
                target.finalize()
        else:
            target = make_vector_target(bfield, result)
            target.begin(len(self.nodes), len(bdry.nodes))
            perform_inverse_index_map(bdry.index_map, target)
            target.finalize()

        return result

    def boundarize_volume_field(self, field, tag=hedge.mesh.TAG_ALL):
        from hedge._internal import perform_index_map
        from hedge.tools import make_vector_target, log_shape

        ls = log_shape(field)
        result = self.boundary_zeros(tag, ls)
        bdry = self._get_boundary(tag)

        if ls != ():
            from pytools import indices_in_shape
            for i in indices_in_shape(ls):
                target = make_vector_target(field[i], result[i])
                target.begin(len(bdry.nodes), len(self.nodes))
                perform_index_map(bdry.index_map, target)
                target.finalize()
        else:
            target = make_vector_target(field, result)
            target.begin(len(bdry.nodes), len(self.nodes))
            perform_index_map(bdry.index_map, target)
            target.finalize()

        return result

    def find_element(self, idx):
        for i, (start, stop) in enumerate(self.element_group):
            if start <= idx < stop:
                return i
        raise ValueError, "not a valid dof index"
        
    def find_face(self, idx):
        el_id = self.find_element(idx)
        el_start, el_stop = self.element_group[el_id]
        for f_id, face_indices in enumerate(self.element_map[el_id].face_indices()):
            if idx-el_start in face_indices:
                return el_id, f_id, idx-el_start
        raise ValueError, "not a valid face dof index"

    # operator binding functions --------------------------------------------------
    @property
    def nabla(self):
        return DiffOperatorVector(self,
                [DifferentiationOperator(self, i) for i in range(self.dimensions)])

    @property
    def minv_stiffness_t(self):
        return DiffOperatorVector(self,
                [MInvSTOperator(self, i) for i in range(self.dimensions)])

    @property
    def stiffness_operator(self):
        return DiffOperatorVector(self,
                [StiffnessOperator(self, i) for i in range(self.dimensions)])

    @property
    def stiffness_t_operator(self):
        return DiffOperatorVector(self,
                [StiffnessTOperator(self, i) for i in range(self.dimensions)])

    @property
    def mass_operator(self):
        return MassOperator(self)

    @property
    def inverse_mass_operator(self):
        return InverseMassOperator(self)

    def get_flux_operator(self, flux, direct=True):
        """Return a flux operator that can be multiplied with
        a volume field to obtain the lifted interior fluxes
        or with a boundary pair to obtain the lifted boundary
        flux.

        `direct' determines whether the operator is applied in a
        matrix-free fashion or uses precomputed matrices.
        """

        from hedge.flux import compile_flux
        from hedge.tools import is_obj_array

        def get_scalar_flux_operator(flux):
            if direct:
                return _DirectFluxOperator(self, flux)
            else:
                return _FluxMatrixOperator(self, flux)

        if isinstance(flux, tuple):
            # a tuple of int/ext fluxes
            return get_scalar_flux_operator([(0,) + flux])
        elif is_obj_array(flux):
            return _VectorFluxOperator(self, 
                    [get_scalar_flux_operator(compile_flux(flux_component)) 
                        for flux_component in flux])
        else:
            return get_scalar_flux_operator(compile_flux(flux))




# random utilities ------------------------------------------------------------
class SymmetryMap(object):
    """A symmetry map on global DG functions.

    Suppose that the L{Mesh} on which a L{Discretization} is defined has
    is mapped onto itself by a nontrivial symmetry map M{f(.)}. Then
    this class allows you to carry out this map on vectors representing
    functions on this L{Discretization}.
    """
    def __init__(self, discr, sym_map, element_map, threshold=1e-13):
        self.discretization = discr

        complete_el_map = {}
        for i, j in element_map.iteritems():
            complete_el_map[i] = j
            complete_el_map[j] = i

        self.map = {}

        for eg in discr.element_groups:
            for el, (el_start, el_stop) in zip(eg.members, eg.ranges):
                mapped_i_el = complete_el_map[el.id]
                mapped_start, mapped_stop = discr.find_el_range(mapped_i_el)
                for i_pt in range(el_start, el_stop):
                    pt = discr.nodes[i_pt]
                    mapped_pt = sym_map(pt)
                    for m_i_pt in range(mapped_start, mapped_stop):
                        if la.norm(discr.nodes[m_i_pt] - mapped_pt) < threshold:
                            self.map[i_pt] = m_i_pt
                            break

                    if i_pt not in self.map:
                        for m_i_pt in range(mapped_start, mapped_stop):
                            print comp.norm_2(discr.nodes[m_i_pt] - mapped_pt)
                        raise RuntimeError, "no symmetry match found"

    def __call__(self, vec):
        result = self.discretization.volume_zeros()
        for i, mapped_i in self.map.iteritems():
            result[mapped_i] = vec[i]
        return result




def generate_random_constant_on_elements(discr):
    result = discr.volume_zeros()
    import random
    for eg in discr.element_groups:
        for e_start, e_end in eg.ranges:
            result[e_start:e_end] = random.random()
    return result




def ones_on_boundary(discr, tag=hedge.mesh.TAG_ALL):
    result = discr.volume_zeros()

    try:
        faces = discr.mesh.tag_to_boundary[tag]
    except KeyError:
        pass
    else:
        for face in faces:
            el, fl = face

            (el_start, el_end), ldis = discr.find_el_data(el.id)
            fl_indices = ldis.face_indices()[fl]

            for i in fl_indices:
                result[el_start+i] = 1

    return result




def ones_on_volume(discr, tag=hedge.mesh.TAG_ALL):
    result = discr.volume_zeros()

    from hedge._internal import UniformElementRanges

    for eg in discr.element_groups:
        if isinstance(eg.ranges, UniformElementRanges):
            result[eg.ranges.start:
                    eg.ranges.start+len(eg.ranges)*eg.ranges.el_size] = 1
        else:
            for e_start, e_end in eg.ranges:
                result[e_start:e_end] = 1

    return result




def integral(discr, volume_vector, tag=hedge.mesh.TAG_ALL):
    try:
        mass_ones = discr._mass_ones
    except AttributeError:
        discr._mass_ones = mass_ones = discr.mass_operator * ones_on_volume(discr, tag)
    
    from hedge.tools import log_shape

    ls = log_shape(volume_vector)
    if ls == ():
        return numpy.dot(mass_ones, volume_vector)
    else:
        result = numpy.zeros(shape=ls, dtype=float)
        
        from pytools import indices_in_shape
        for i in indices_in_shape(ls):
            result[i] = numpy.dot(mass_ones, volume_vector[i])
            #result[i] = (discr.mass_operator*volume_vector[i]).sum()

        return result




def norm(discr, volume_vector, p=2):
    if p == numpy.Inf:
        return numpy.abs(volume_vector).max()
    else:
        from hedge.tools import log_shape

        if p != 2:
            volume_vector = numpy.abs(volume_vector)**(p/2)

        ls = log_shape(volume_vector)
        if ls == ():
            return float(numpy.dot(
                    volume_vector,
                    discr.mass_operator * volume_vector)**(1/p))
        else:
            assert len(ls) == 1
            return float(sum(
                    numpy.dot(
                        subv,
                        discr.mass_operator * subv)
                    for subv in volume_vector)**(1/p))

    

# operator wrapper ------------------------------------------------------------
class OpWrapper(hedge.tools.OperatorBase):
    def __init__(self,  discr_op):
        operator.Operator(num.Float64).__init__(self)
        self.discr_op = discr_op

    def size1(self):
        return len(self.discr_op.discr)

    def size2(self):
        return len(self.discr_op.discr)

    def apply(self, before, after):
        after[:] = self.discr_op.__mul__(before)





# operator algebra ------------------------------------------------------------
class DiscretizationVectorOperator(object):
    def __init__(self, discr):
        self.discr = discr

    def __add__(self, op2):
        return _OperatorSum(self, op2)

    def __sub__(self, op2):
        return _OperatorSum(self, -op2)

    def __neg__(self):
        return _ScalarMultipleOperator(-1, self)

    def __rmul__(self, scalar):
        value = float(scalar)
        return _ScalarMultipleOperator(scalar, self)

class _ScalarMultipleOperator(DiscretizationVectorOperator):
    def __init__(self, scalar, op):
        self.scalar = scalar
        self.op = op

    def __mul__(self, field):
        return self.scalar*(self.op*field)

class _OperatorSum(DiscretizationVectorOperator):
    def __init__(self, op1, op2):
        self.op1 = op1
        self.op2 = op2

    def __mul__(self, field):
        return self.op1*field + self.op2*field




# diff operators --------------------------------------------------------------
class _DiffResultsCache(object):
    def __init__(self, vector):
        self.vector = vector
        self.cache = {}

    def lookup(self, diff_op):
        return self.cache.get(diff_op, None)

    def set(self, diff_op, value):
        self.cache[diff_op] = value




def cache_diff_results(vec):
    from hedge.tools import log_shape
    ls = log_shape(vec)

    if ls == ():
        return _DiffResultsCache(vec)
    else:
        result = numpy.empty(ls, dtype=object)
        from pytools import indices_in_shape
        for i in indices_in_shape(ls):
            result[i] = _DiffResultsCache(vec[i])
        return result




class DiffOperatorBase(DiscretizationVectorOperator):
    def __init__(self, discr, xyz_axis):
        DiscretizationVectorOperator.__init__(self, discr)

        self.xyz_axis = xyz_axis

        self.do_warn = True

        if self.discr.instrumented:
            from pytools.log import time_and_count_function
            self.__mul__ = \
                    time_and_count_function(
                            self.__mul__,
                            self.discr.diff_op_timer,
                            self.discr.diff_op_counter)

    def do_not_warn(self):
        self.do_warn = False

    def diff_rst(self, rst_axis, field):
        result = self.discr.volume_zeros()

        from hedge.tools import make_vector_target
        target = make_vector_target(field, result)

        target.begin(len(self.discr), len(self.discr))

        from hedge._internal import perform_elwise_operator
        for eg in self.discr.element_groups:
            perform_elwise_operator(eg.ranges, eg.ranges, 
                    self.matrices(eg)[rst_axis], target)

        target.finalize()

        return result

    def diff_xyz(self, field, result):
        if isinstance(field, _DiffResultsCache):
            cache = field
            field = cache.vector
            rst_derivatives = cache.lookup(self.__class__)
        else:
            cache = None
            rst_derivatives = None
            if self.do_warn:
                from warnings import warn
                warn("wrap operand of differential operator in cache_diff_results() for speed")

        if rst_derivatives is None:
            rst_derivatives = [self.diff_rst(i, field) 
                    for i in range(self.discr.dimensions)]
            if cache:
                cache.set(self.__class__, rst_derivatives)

        from hedge.tools import make_vector_target
        from hedge._internal import perform_elwise_scale

        for rst_axis in range(self.discr.dimensions):
            target = make_vector_target(rst_derivatives[rst_axis], result)

            target.begin(len(self.discr), len(self.discr))
            for eg in self.discr.element_groups:
                perform_elwise_scale(eg.ranges,
                        self.coefficients(eg)[self.xyz_axis][rst_axis],
                        target)
            target.finalize()
        return result

    def __mul__(self, field):
        from hedge.tools import log_shape
        lshape = log_shape(field)

        result = self.discr.volume_zeros(lshape)

        if lshape == ():
            from pytools import indices_in_shape
            self.diff_xyz(field, result)
        else:
            from pytools import indices_in_shape
            for i in indices_in_shape(lshape):
                self.diff_xyz(field[i], result[i])

        return result
            
    def perform_on(self, target):
        from hedge._internal import perform_elwise_scaled_operator

        target.begin(len(self.discr), len(self.discr))

        for eg in self.discr.element_groups:
            for coeff, mat in zip(self.coefficients(eg)[self.xyz_axis], 
                    self.matrices(eg)):
                perform_elwise_scaled_operator(
                        eg.ranges, eg.ranges, coeff, mat, target)

        target.finalize()

class DifferentiationOperator(DiffOperatorBase):
    @staticmethod
    def matrices(element_group): return element_group.differentiation_matrices

    @staticmethod
    def coefficients(element_group): return element_group.diff_coefficients

class MInvSTOperator(DiffOperatorBase):
    @staticmethod
    def matrices(element_group): return element_group.minv_st

    @staticmethod
    def coefficients(element_group): return element_group.diff_coefficients

class StiffnessOperator(DiffOperatorBase):
    @staticmethod
    def matrices(element_group): return element_group.stiffness_matrices

    @staticmethod
    def coefficients(element_group): return element_group.stiffness_coefficients

class StiffnessTOperator(DiffOperatorBase):
    @staticmethod
    def matrices(element_group): return element_group.stiffness_t_matrices

    @staticmethod
    def coefficients(element_group): return element_group.stiffness_coefficients

class DiffOperatorVector(object):
    def __init__(self, discr, operators):
        self.discr = discr
        self.operators = operators

    def __len__(self):
        return len(self.operators)

    def __getitem__(self, i):
        return self.operators[i]

    def __mul__(self, field):
        result = self.discr.volume_zeros(shape=(len(self),))
        for i, op in enumerate(self):
            op.diff_xyz(field, result[i])
        return result




# mass operators --------------------------------------------------------------
class MassOperatorBase(DiscretizationVectorOperator):
    def __init__(self, discr):
        DiscretizationVectorOperator.__init__(self, discr)

        if self.discr.instrumented:
            from pytools.log import time_and_count_function
            self.perform_on = \
                    time_and_count_function(
                            self.perform_on,
                            self.discr.mass_op_timer,
                            self.discr.mass_op_counter)


    def __mul__(self, field):
        from hedge.tools import log_shape, make_vector_target
        lshape = log_shape(field)
        result = self.discr.volume_zeros(lshape)

        from pytools import indices_in_shape
        for i in indices_in_shape(lshape):
            self.perform_on(make_vector_target(field[i], result[i]))

        return result

    def perform_on(self, target):
        from hedge._internal import perform_elwise_scaled_operator
        target.begin(len(self.discr), len(self.discr))
        for eg in self.discr.element_groups:
            perform_elwise_scaled_operator(eg.ranges, eg.ranges,
                   self.coefficients(eg), self.matrix(eg), 
                   target)
        target.finalize()

class MassOperator(MassOperatorBase):
    @staticmethod
    def matrix(element_group): return element_group.mass_matrix

    @staticmethod
    def coefficients(element_group): return element_group.jacobians

class InverseMassOperator(MassOperatorBase):
    @staticmethod
    def matrix(element_group): return element_group.inverse_mass_matrix

    @staticmethod
    def coefficients(element_group): return element_group.inverse_jacobians




# flux operators --------------------------------------------------------------
class _DirectFluxOperator(DiscretizationVectorOperator):
    def __init__(self, discr, flux):
        DiscretizationVectorOperator.__init__(self, discr)
        self.flux = flux

    def mul_to(self, field, out):
        from hedge.tools import make_vector_target, log_shape

        if isinstance(field, BoundaryPair):
            # boundary flux
            bpair = field
            bdry = self.discr._get_boundary(bpair.tag)

            if not bdry.nodes:
                return 0

            field = bpair.field
            bfield = bpair.bfield

            class ZeroVector:
                dtype = 0
                def __getitem__(self, idx):
                    return 0

            if log_shape(field) != () or log_shape(field) != ():
                if isinstance(bfield, int) and bfield == 0:
                    bfield = ZeroVector()
                if isinstance(field, int) and field == 0:
                    field = ZeroVector()

                for idx, int_flux, ext_flux in self.flux:
                    if not (0 <= idx < len(field)):
                        raise RuntimeError, "flux depends on out-of-bounds field index"
                    self.discr._perform_boundary_flux(
                            int_flux, make_vector_target(field[idx], out),
                            ext_flux, make_vector_target(bfield[idx], out), 
                            bdry)
            else:
                if len(self.flux) == 1:
                    idx, int_flux, ext_flux = self.flux[0]
                    if idx != 0:
                        raise RuntimeError, "flux depends on out-of-bounds field index"
                    self.discr._perform_boundary_flux(
                            int_flux, make_vector_target(field, out),
                            ext_flux, make_vector_target(bfield, out), 
                            bdry)
                elif len(self.flux) > 1:
                    raise RuntimeError, "only found one field to process, but flux has multiple field dependencies"

            return out

        else:
            # inner flux
            from hedge.tools import log_shape
            ls = log_shape(field)
            if ls != ():
                assert len(ls) == 1
                for idx, int_flux, ext_flux in self.flux:
                    if not (0 <= idx < len(field)):
                        raise RuntimeError, "flux depends on out-of-bounds field index"
                    self.discr.perform_inner_flux(
                            int_flux, ext_flux, make_vector_target(field[idx], out))
            else:
                if len(self.flux) > 1:
                    raise RuntimeError, "only found one field to process, but flux has multiple field dependencies"

                idx, int_flux, ext_flux = self.flux[0]
                if idx != 0:
                    raise RuntimeError, "flux depends on out-of-bounds field index"
                self.discr.perform_inner_flux(
                        int_flux, ext_flux, make_vector_target(field, out))

            return out

    def __mul__(self, field):
        result = self.discr.volume_zeros()
        return self.mul_to(field, result)

    def perform_inner(self, tgt):
        dof = len(self.discr)

        for idx, int_flux, ext_flux in self.flux:
            self.discr.perform_inner_flux(int_flux, ext_flux, 
                    tgt.rebased_target(0, dof*idx))

    def perform_int_bdry(self, tag, tgt):
        from hedge._internal import NullTarget
        dof = len(self.discr)

        for idx, int_flux, ext_flux in self.flux:
            self.discr._perform_boundary_flux(
                    int_flux, tgt.rebased_target(0, dof*idx),
                    ext_flux, NullTarget(), 
                    self.discr._get_boundary(tag))




class _FluxMatrixOperator(DiscretizationVectorOperator):
    def __init__(self, discr, flux):
        DiscretizationVectorOperator.__init__(self, discr)

        from hedge._internal import MatrixTarget
        self.flux = flux

        dof = len(self.discr)

        self.inner_matrices = {}
        for idx, int_flux, ext_flux in self.flux:
            inner_bmatrix = pyublas.zeros(shape=(dof, dof), flavor=pyublas.SparseBuildMatrix)
            discr.perform_inner_flux(int_flux, ext_flux, MatrixTarget(inner_bmatrix))
            self.inner_matrices[idx] = \
                    pyublas.asarray(inner_bmatrix, flavor=pyublas.SparseExecuteMatrix)

        self.bdry_ops = {}

    def __mul__(self, field):

        def get_bdry_op(idx, bdry, int_flux, ext_flux):
            try:
                return self.bdry_ops[idx, bdry]
            except KeyError:
                dof = len(self.discr)
                bdry_dof = len(bdry.nodes)

                from hedge._internal import MatrixTarget

                int_bmatrix = pyublas.zeros(shape=(dof,dof), flavor=pyublas.SparseBuildMatrix)
                ext_bmatrix = pyublas.zeros(shape=(dof,bdry_dof), flavor=pyublas.SparseBuildMatrix)
                self.discr._perform_boundary_flux(
                        int_flux, MatrixTarget(int_bmatrix), 
                        ext_flux, MatrixTarget(ext_bmatrix), 
                        bdry)
                int_matrix = pyublas.asarray(int_bmatrix, flavor=pyublas.SparseExecuteMatrix)
                ext_matrix = pyublas.asarray(ext_bmatrix, flavor=pyublas.SparseExecuteMatrix)

                self.bdry_ops[idx, bdry] = int_matrix, ext_matrix

                return int_matrix, ext_matrix

        if isinstance(field, BoundaryPair):
            # boundary flux
            bpair = field
            bdry = self.discr._get_boundary(bpair.tag)

            if not bdry.nodes:
                return 0

            class ZeroVector:
                def __getitem__(self, idx):
                    return 0

            field = bpair.field
            bfield = bpair.bfield

            from hedge.tools import log_shape
            ls_f = log_shape(field)
            ls_bf = log_shape(bfield)

            if ls_f != () or ls_bf != ():
                if isinstance(bfield, int) and bfield == 0:
                    bfield = ZeroVector()
                elif isinstance(field, int) and field == 0:
                    field = ZeroVector()

                assert len(ls) == 1
                result = self.discr.volume_zeros()
                for idx, int_flux, ext_flux in self.flux:
                    int_matrix, ext_matrix = get_bdry_op(idx, bdry, int_flux, ext_flux)
                    result += int_matrix*bpair.field + ext_matrix*bpair.bfield
                return result
            else:
                assert len(self.flux) == 1
                idx, int_flux, ext_flux = self.flux[0]
                assert idx == 0

                int_matrix, ext_matrix = get_bdry_op(0, bdry, int_flux, ext_flux)
                return int_matrix*bpair.field + ext_matrix*bpair.bfield
        else:
            from hedge.tools import log_shape

            ls = log_shape(field)
            if ls != ():
                assert len(ls) == 1
                result = self.discr.volume_zeros()
                for idx, int_flux, ext_flux in self.flux:
                    result += self.inner_matrices[idx] * field[idx]
                return result
            else:
                assert len(self.flux) == 1
                idx, int_flux, ext_flux = self.flux[0]
                assert idx == 0
                return self.inner_matrices[0] * field

    def matrix_inner(self):
        """Returns a BlockMatrix to compute the lifting of the domain-interior fluxes.

        This matrix can only capture the interior part of the flux, the exterior-facing
        part is not taken into account.

        The different components are assumed to be run together in one vector,
        in order.
        """
        from hedge.tools import BlockMatrix
        flen = len(self.discr.volume_zeros())
        return BlockMatrix(
                (0, flen*idx, self.inner_matrices[idx])
                for idx, int_flux, ext_flux in self.flux)




class _VectorFluxOperator(object):
    def __init__(self, discr, flux_operators):
        self.discr = discr
        self.flux_operators = flux_operators

        if discr.instrumented:
            from pytools.log import time_and_count_function
            self._do_inner_flux = \
                    time_and_count_function(
                            self._do_inner_flux,
                            self.discr.inner_flux_timer,
                            self.discr.inner_flux_counter,
                            increment=sum(len(fo) for fo in self.flux_operators))

    def __mul__(self, field):
        if isinstance(field, BoundaryPair):
            result = self.discr.volume_zeros(
                    shape=(len(self.flux_operators),))

            for i, fop_i in enumerate(self.flux_operators):
                fop_i.mul_to(field, result[i])

            return result
        else:
            return self._do_inner_flux(field)

    def _do_inner_flux(self, field):
        # this is for performance -- it is faster to apply several fluxes
        # to a single operand at once
        result = self.discr.volume_zeros(
                shape=(len(self.flux_operators),))

        def find_field_flux(flux_op, i_field):
            for idx, int_flux, ext_flux in flux_op.flux:
                if idx == i_field:
                    return int_flux, ext_flux
            return None

        from hedge.tools import log_shape
        if log_shape(field) == ():
            field = field[numpy.newaxis,:]

        from hedge._internal import \
                perform_multiple_double_sided_fluxes_on_single_operand, \
                ChainedFlux
        for i_field, f_i in enumerate(field):
            fluxes_and_results = []
            for i_result, fo in enumerate(self.flux_operators):
                scalar_flux = find_field_flux(fo, i_field)
                if scalar_flux is not None:
                    int_flux, ext_flux = scalar_flux
                    fluxes_and_results.append(
                            (ChainedFlux(int_flux), 
                                ChainedFlux(ext_flux), 
                                result[i_result]))
            for fg, fmm in self.discr.face_groups:
                perform_multiple_double_sided_fluxes_on_single_operand(
                        fg, fmm, fluxes_and_results, f_i)

        return result






    def perform_inner(self, tgt):
        dof = len(self.discr)
        for i, fo in enumerate(self.flux_operators):
            fo.perform_inner(tgt.rebased_target(dof*i, 0))

    def perform_int_bdry(self, tag, tgt):
        dof = len(self.discr)
        for i, fo in enumerate(self.flux_operators):
            fo.perform_int_bdry(tag, tgt.rebased_target(dof*i, 0))




# boundary treatment ----------------------------------------------------------
class BoundaryPair(object):
    """Represents a pairing of a volume and a boundary field, used for the
    application of boundary fluxes.

    Do not use class constructor. Use L{pair_with_boundary}() instead.
    """
    def __init__(self, field, bfield, tag):
        self.field = field
        self.bfield = bfield
        self.tag = tag




def pair_with_boundary(field, bfield, tag=hedge.mesh.TAG_ALL):
    """Create a L{BoundaryPair} out of the volume field C{field}
    and the boundary field C{field}.

    Also accepts ArithmeticList or 0 for either argument.
    """
    return BoundaryPair(field, bfield, tag)




# projection between different discretizations --------------------------------
class Projector:
    def __init__(self, from_discr, to_discr):
        self.from_discr = from_discr
        self.to_discr = to_discr

        self.interp_matrices = []
        for from_eg, to_eg in zip(
                from_discr.element_groups, to_discr.element_groups):
            from_ldis = from_eg.local_discretization
            to_ldis = to_eg.local_discretization

            from_count = from_ldis.node_count()
            to_count = to_ldis.node_count()

            # check that the two element groups have the same members
            for from_el, to_el in zip(from_eg.members, to_eg.members):
                assert from_el is to_el

            from hedge.tools import permutation_matrix

            # assemble the from->to mode permutation matrix, guided by 
            # mode identifiers
            if to_count > from_count:
                to_node_ids_to_idx = dict(
                        (nid, i) for i, nid in 
                        enumerate(to_ldis.generate_mode_identifiers()))

                to_indices = [
                    to_node_ids_to_idx[from_nid]
                    for from_nid in from_ldis.generate_mode_identifiers()
                    ]

                pmat = permutation_matrix(
                    to_indices=to_indices, 
                    h=to_count, w=from_count)
            else:
                from_node_ids_to_idx = dict(
                        (nid, i) for i, nid in 
                        enumerate(from_ldis.generate_mode_identifiers()))

                from_indices = [
                    from_node_ids_to_idx[to_nid]
                    for to_nid in to_ldis.generate_mode_identifiers()
                    ]

                pmat = permutation_matrix(
                    from_indices=from_indices, 
                    h=to_count, w=from_count)

            # build interpolation matrix
            from_matrix = from_ldis.vandermonde()
            to_matrix = to_ldis.vandermonde()

            from hedge.tools import leftsolve
            from numpy import dot
            self.interp_matrices.append(
                    numpy.asarray(
                        leftsolve(from_matrix, dot(to_matrix, pmat)),
                        order="C"))

    def __call__(self, from_vec):
        from hedge._internal import perform_elwise_operator, VectorTarget
        from hedge.tools import log_shape

        ls = log_shape(from_vec)
        result = self.to_discr.volume_zeros(ls)

        from pytools import indices_in_shape
        for i in indices_in_shape(ls):
            target = VectorTarget(from_vec[i], result[i])

            target.begin(len(self.to_discr), len(self.from_discr))
            for from_eg, to_eg, imat in zip(
                    self.from_discr.element_groups, 
                    self.to_discr.element_groups, 
                    self.interp_matrices):
                perform_elwise_operator(
                        from_eg.ranges, to_eg.ranges, 
                        imat, target)

            target.finalize()

        return result




# filter ----------------------------------------------------------------------
class ExponentialFilterResponseFunction:
    """A typical exponential-falloff mode response filter function.

    See description in Section 5.6.1 of Jacobs/Hesthaven.
    """
    def __init__(self, min_amplification=0.1, order=6):
        """Construct the filter function.

        @arg min_amplification: The amplification factor applied to the highest mode.
        @arg order: The order of the filter. This controls how fast (or slow) the
          C{min_amplification} is reached.

        The amplification factor of the lowest-order (constant) mode is always 1.
        """
        from math import log
        self.alpha = -log(min_amplification)
        self.order = order

    def __call__(self, mode_idx, ldis):
        eta = max(mode_idx)/ldis.order

        from math import exp
        return exp(-self.alpha * eta**self.order)




class Filter:
    def __init__(self, discr, mode_response_func):
        """Construct a filter.

        @arg discr: The L{Discretization} for which the filter is to be
          constructed.
        @mode_response_func: A function mapping 
          C{(mode_tuple, local_discretization)} to a float indicating the
          factor by which this mode is to be multiplied after filtering.
        """
        self.discr = discr

        self.filter_matrices = []
        for eg in discr.element_groups:
            ldis = eg.local_discretization

            node_count = ldis.node_count()

            filter_coeffs = [mode_response_func(mid, ldis)
                for mid in ldis.generate_mode_identifiers()] 

            # build filter matrix
            vdm = ldis.vandermonde()
            from hedge.tools import leftsolve
            from numpy import dot
            self.filter_matrices.append(numpy.asarray(
                leftsolve(vdm,
                    dot(vdm, numpy.diag(filter_coeffs))),
                order="C"))

    def __call__(self, vec):
        from hedge.tools import log_shape

        ls = log_shape(vec)
        result = self.discr.volume_zeros(ls)

        from pytools import indices_in_shape
        for i in indices_in_shape(ls):
            from hedge._internal import perform_elwise_operator, VectorTarget

            target = VectorTarget(vec[i], result[i])

            target.begin(len(self.discr), len(self.discr))
            for eg, fmat in zip(self.discr.element_groups, self.filter_matrices):
                perform_elwise_operator(eg.ranges, eg.ranges, fmat, target)

            target.finalize()

        return result



def plot_order_3_tri_poly_components():
    from hedge.polynomial import jacobi_function
    from hedge.tools import plot_1d

    i = 3
    j = 3
    f = jacobi_function(0, 0, i)
    g = jacobi_function(2*i+1, 0, j)
    plot_1d(f, -1, 1)
    plot_1d(g, -1, 1)

def plot_tri_polynomials_hi_res():
    import numpy
    import pylinear.array as num
    from hedge.element import TriangularElement
    from hedge.polynomial import generic_vandermonde
    from pyvtk import PolyData, PointData, VtkData, Scalars, Vectors

    tri_low = TriangularElement(3)
    tri_high = TriangularElement(14)

    unodes = tri_high.unit_nodes()

    low_vdm = tri_low.vandermonde()
    point_values = generic_vandermonde(unodes, tri_low.basis_functions()).T
    lagrange_values = (1/low_vdm.T) * point_values

    def three_vector(x):
        if len(x) == 3:
            return x
        elif len(x) == 2:
            return x[0], x[1], 0.
        elif len(x) == 1:
            return x[0], 0, 0.

    structure = PolyData(points=[three_vector(x) for x in unodes], 
            polygons=tri_high.generate_submesh_indices())
    pdatalist = []
    for i, row in enumerate(lagrange_values):
        pdatalist.append(Scalars(numpy.array(row), 
            name="basis%d" % i, lookup_table="default"))
    vtk = VtkData(structure, "Hedge visualization", PointData(*pdatalist))
    vtk.tofile("basis.vtk")

def plot_tri_quad_weights():
    import numpy
    import pylinear.array as num
    from hedge.element import TriangularElement
    from hedge.polynomial import generic_vandermonde
    from pyvtk import PolyData, PointData, VtkData, Scalars, Vectors

    edata = TriangularElement(3)
    unodes = edata.unit_nodes()
    mass_mat = edata.mass_matrix()
    ones = num.ones((mass_mat.shape[0],))

    point_values = mass_mat<<num.solve>>ones
    #point_values = mass_mat*ones

    def three_vector(x):
        if len(x) == 3:
            return x
        elif len(x) == 2:
            return x[0], x[1], 0.
        elif len(x) == 1:
            return x[0], 0, 0.

    structure = PolyData(points=[three_vector(x) for x in unodes], 
            polygons=edata.generate_submesh_indices())
    pdatalist = [
           Scalars(numpy.array(point_values), 
            name="qweights", lookup_table="default") ]
    vtk = VtkData(structure, "Hedge visualization", PointData(*pdatalist))
    vtk.tofile("quadweights.vtk")

def plot_real_tri_quad_weights():
    from hedge.cubature import TetrahedronCubatureData
    outf = file("quad.dat", "w")
    for x, y, w in TetrahedronCubatureData[15]:
        outf.write("%g\t%g\t%g\n" % (x,y,w))

def visualize_single_tet():
    from hedge.mesh import ConformalMesh
    from hedge.element import TetrahedralElement
    from hedge.discretization import Discretization
    from hedge.visualization import SiloVisualizer

    points = [
            (-1,-1,-1),
            (+1,-1,-1),
            (-1,+1,-1),
            (-1,-1,+1),
            ]
    elements = [(0,1,2,3)]
    mesh = ConformalMesh(points, elements)
    discr = Discretization(mesh, TetrahedralElement(2))
    vis = SiloVisualizer(discr)
    vis("single-tet.silo", [("zip", discr.volume_zeros())])



if __name__ == "__main__":
    #plot_tri_order_1_polynomials_hi_res()
    #plot_tri_quad_weights()
    #plot_real_tri_quad_weights()
    visualize_single_tet()



# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




import pylinear.array as num
import pylinear.computation as comp
from pytools.arithmetic_container import ArithmeticList
from hedge.tools import Rotation, dot




class StrongWaveOperator:
    def __init__(self, discr, source_f=None):
        self.discr = discr
        self.source_f = source_f

        from hedge.flux import zero, trace_sign, make_normal, local, neighbor, average
        from hedge.discretization import bind_flux, bind_nabla, bind_mass_matrix, \
                bind_inverse_mass_matrix

        normal = make_normal(discr.dimensions)
        flux_weak = average*normal
        flux_strong = local*normal - flux_weak

        self.nabla = bind_nabla(discr)
        self.mass = bind_mass_matrix(discr)
        self.m_inv = bind_inverse_mass_matrix(discr)

        self.flux = bind_flux(discr, flux_strong)

    def rhs(self, t, y):
        from hedge.discretization import pair_with_boundary

        u = y[0]
        v = y[1:]

        #bc_v = self.discr.boundarize_volume_field(v)
        bc_u = -self.discr.boundarize_volume_field(u)

        rhs = ArithmeticList([])
        # rhs u
        rhs.append(dot(self.nabla, v) 
                - self.m_inv*(
                    dot(self.flux, v) 
                    #+ dot(bflux, pair_with_boundary(v, bc_v))
                    ))

        if self.source_f is not None:
            rhs[0] += self.source_f(t)

        # rhs v
        #rhs.extend(self.nabla*u 
                #-self.m_inv*(
                    #self.flux*u 
                    #+ self.flux*pair_with_boundary(u, bc_u)
                    #))
        #rhs.extend( -(self.m_inv*( self.flux*u )))
        #rhs.extend(self.nabla*u)
        rhs.extend(self.flux*pair_with_boundary(u, bc_u))
        return rhs




def main() :
    from hedge.element import \
            TriangularElement, \
            TetrahedralElement
    from hedge.timestep import RK4TimeStepper
    from hedge.mesh import \
            make_disk_mesh, \
            make_regular_square_mesh, \
            make_square_mesh, \
            make_ball_mesh
    from pylo import DB_VARTYPE_VECTOR
    from hedge.visualization import \
            SiloVisualizer, \
            make_silo_file, \
            get_rank_partition
    from hedge.discretization import Discretization
    from pytools.stopwatch import Job
    from math import sin, cos, pi, exp, sqrt
    from hedge.parallel import \
            guess_parallelization_context, \
            reassemble_volume_field

    pcon = guess_parallelization_context()

    dim = 2
    order = 3

    if dim == 2:
        if pcon.is_head_rank:
            mesh = make_disk_mesh()
        #mesh = make_regular_square_mesh(n=5)
        #mesh = make_square_mesh(max_area=0.008)
        #mesh.transform(Rotation(pi/8))
        el_class = TriangularElement
    elif dim == 3:
        if pcon.is_head_rank:
            mesh = make_ball_mesh(max_volume=0.001)
        el_class = TetrahedralElement
    else:
        raise RuntimeError, "bad number of dimensions"

    if pcon.is_head_rank:
        print "%d elements" % len(mesh.elements)
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()

    discr = pcon.make_discretization(mesh_data, el_class(order))
    stepper = RK4TimeStepper()
    vis = SiloVisualizer(discr)

    dt = discr.dt_factor(1)
    nsteps = int(1/dt)
    if pcon.is_head_rank:
        print "dt", dt
        print "nsteps", nsteps

    def ic_u(x):
        return exp(-x*x*128)

    op = StrongWaveOperator(discr)
    fields = ArithmeticList(
            [discr.interpolate_volume_function(ic_u)])
    fields.extend(op.nabla*fields[0])

    if pcon.is_head_rank:
        gdiscr = Discretization(mesh, el_class(order))
        gvis = SiloVisualizer(gdiscr)
        gstepper = RK4TimeStepper()
        gop = StrongWaveOperator(gdiscr)

        gfields = ArithmeticList(
                [gdiscr.interpolate_volume_function(ic_u)])
        gfields.extend(gop.nabla*gfields[0])
    else:
        gdiscr = None

    gpart = reassemble_volume_field(pcon, gdiscr, discr, 
            get_rank_partition(pcon, discr))

    for step in range(nsteps):
        t = step*dt
        if step % 1 == 0:
            print "timestep %d, t=%f, l2=%g" % (
                    step, t, sqrt(fields[0]*(op.mass*fields[0])))

        if True:
            rhs = op.rhs(t, fields)
            g_ass_rhs = reassemble_volume_field(pcon, gdiscr, discr, rhs)
            if pcon.is_head_rank:
                grhs = gop.rhs(t, gfields)
                silo = make_silo_file("rhs-%04d" % step)
                gvis.add_to_silo(silo,
                        scalars=[
                            ("u", gfields[0]), 
                            ("par_rhs_u", g_ass_rhs[0]), 
                            ("ser_rhs_u", grhs[0]), 
                            ("partition", gpart),
                            ],
                        vectors=[
                            ("v", gfields[1:]), 
                            ("par_rhs_v", g_ass_rhs[1:]), 
                            ("ser_rhs_v", grhs[1:]), 
                            ],
                        expressions=[
                            ("d_rhs_v", "par_rhs_v-ser_rhs_v", 
                                DB_VARTYPE_VECTOR),
                            ("d_rhs_u", "par_rhs_u-ser_rhs_u"),
                            ]
                            )
                silo.close()

        if False:
            silo = make_silo_file("fld-%04d" % step, pcon)
            vis.add_to_silo(silo,
                    [("u", fields[0]), ], 
                    [("v", fields[1:]), ],
                    time=t,
                    step=step)
            silo.close()

        fields = stepper(fields, t, dt, op.rhs)
        if pcon.is_head_rank:
            gfields = gstepper(gfields, t, dt, gop.rhs)

if __name__ == "__main__":
    #import cProfile as profile
    #profile.run("main()", "wave2d.prof")
    main()




# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




from __future__ import division
import pylinear.array as num
import pylinear.computation as comp




def main():
    from hedge.element import TriangularElement
    from hedge.timestep import RK4TimeStepper
    from hedge.mesh import make_disk_mesh
    from hedge.visualization import \
            VtkVisualizer, \
            SiloVisualizer, \
            get_rank_partition
    from pylo import DB_VARTYPE_VECTOR
    from hedge.tools import dot, EOCRecorder
    from math import sqrt, pi, exp
    from analytic_solutions import \
            check_time_harmonic_solution, \
            RealPartAdapter, \
            SplitComplexAdapter, \
            CartesianAdapter, \
            CylindricalCavityMode, \
            RectangularWaveguideMode, \
            RectangularCavityMode
    from hedge.operators import TEMaxwellOperator, TMMaxwellOperator
    from hedge.parallel import guess_parallelization_context
    from hedge.data import GivenFunction, TimeIntervalGivenFunction
    from pytools.arithmetic_container import ArithmeticList

    pcon = guess_parallelization_context()

    epsilon0 = 8.8541878176e-12 # C**2 / (N m**2)
    mu0 = 4*pi*1e-7 # N/A**2.
    epsilon = 1*epsilon0
    mu = 1*mu0

    #eoc_rec = EOCRecorder()

    cylindrical = False
    periodic = False

    mesh = make_disk_mesh(r=0.5)

    if pcon.is_head_rank:
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()

    class CurrentSource:
        shape = (3,)

        def __call__(self, x):
            return [0,0,exp(-80*comp.norm_2_squared(x))]

    #for order in [1,2,3,4,5,6]:
    for order in [3]:
        discr = pcon.make_discretization(mesh_data, TriangularElement(order))

        vis = VtkVisualizer(discr, pcon, "em-%d" % order)
        #vis = SiloVisualizer(discr, pcon)

        dt = discr.dt_factor(1/sqrt(mu*epsilon))
        final_time = dt*200
        nsteps = int(final_time/dt)+1
        dt = final_time/nsteps

        if pcon.is_head_rank:
            print "---------------------------------------------"
            print "order %d" % order
            print "---------------------------------------------"
            print "dt", dt
            print "nsteps", nsteps
            print "#elements=", len(mesh.elements)

        def l2_norm(field):
            return sqrt(dot(field, discr.mass_operator*field))

        op = TMMaxwellOperator(discr, epsilon, mu, upwind_alpha=1,
                direct_flux=True,
                current=TimeIntervalGivenFunction(
                    GivenFunction(CurrentSource()), off_time=final_time/10)
                )
        fields = op.assemble_fields()

        stepper = RK4TimeStepper()
        from time import time
        last_tstep = time()
        t = 0
        for step in range(nsteps):
            e, h = op.split_eh(fields)
            print "timestep %d, t=%g l2[e]=%g l2[h]=%g secs=%f" % (
                    step, t, l2_norm(e), l2_norm(h),
                    time()-last_tstep)
            last_tstep = time()

            if True:
                visf = vis.make_file("em-%d-%04d" % (order, step))
                vis.add_data(visf,
                        [ ("e", e), ("h", h), ],
                        time=t, step=step
                        )
                visf.close()

            fields = stepper(fields, t, dt, op.rhs)
            t += dt

        #true_fields = discr.interpolate_volume_function(r_sol)
        #eoc_rec.add_data_point(order, l2_norm(fields-true_fields))

        #print
        #print eoc_rec.pretty_print("P.Deg.", "L2 Error")

if __name__ == "__main__":
    import cProfile as profile
    #profile.run("main()", "wave2d.prof")
    main()




"""Canned operators for several PDEs, such as Maxwell's, heat, Poisson, etc."""

from __future__ import division

__copyright__ = "Copyright (C) 2007 Andreas Kloeckner"

__license__ = """
This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see U{http://www.gnu.org/licenses/}.
"""





import numpy
import numpy.linalg as la
import pyublas
import hedge.tools
import hedge.mesh
import hedge.data




class Operator(object):
    """A base class for Discontinuous Galerkin operators.

    You may derive your own operators from this class, but, at present
    this class provides no functionality. Its function is merely as 
    documentation, to group related classes together in an inheritance
    tree.
    """
    pass




class TimeDependentOperator(Operator):
    """A base class for time-dependent Discontinuous Galerkin operators.

    You may derive your own operators from this class, but, at present
    this class provides no functionality. Its function is merely as 
    documentation, to group related classes together in an inheritance
    tree.
    """
    pass




class GradientOperator(Operator):
    def __init__(self, discr):
        self.discr = discr

        from hedge.flux import make_normal, FluxScalarPlaceholder
        u = FluxScalarPlaceholder()

        self.nabla = discr.nabla
        self.m_inv = discr.inverse_mass_operator
        normal = make_normal(self.discr.dimensions)
        self.flux = discr.get_flux_operator(u.int*normal - u.avg*normal)

    def __call__(self, u):
        from hedge.mesh import TAG_ALL
        from hedge.discretization import pair_with_boundary, cache_diff_results

        bc = self.discr.boundarize_volume_field(u, TAG_ALL)

        return self.nabla*cache_diff_results(u) - self.m_inv*(
                self.flux * u + 
                self.flux * pair_with_boundary(u, bc, TAG_ALL))




class DivergenceOperator(Operator):
    def __init__(self, discr, subset=None):
        self.discr = discr

        if subset is None:
            subset = self.subset = discr.dimensions * [True,]
        else:
            # chop off any extra dimensions
            subset = self.subset = subset[:discr.dimensions]

        from hedge.flux import make_normal, FluxVectorPlaceholder
        v = FluxVectorPlaceholder(discr.dimensions)

        self.nabla = discr.nabla
        self.m_inv = discr.inverse_mass_operator
        normal = make_normal(self.discr.dimensions)

        flux = 0
        for i, i_enabled in enumerate(subset):
            if i_enabled:
                flux += (v.int-v.avg)[i]*normal[i]

        self.flux = discr.get_flux_operator(flux)

    def __call__(self, v):
        from hedge.mesh import TAG_ALL
        from hedge.discretization import pair_with_boundary, cache_diff_results

        bc = self.discr.boundarize_volume_field(v, TAG_ALL)

        v_cache = cache_diff_results(v)
        local_op_result = self.discr.volume_zeros()
        for i, i_enabled in enumerate(self.subset):
            if i_enabled:
                local_op_result += self.nabla[i]*v_cache[i]
        
        return local_op_result - self.m_inv*(
                self.flux * v + 
                self.flux * pair_with_boundary(v, bc, TAG_ALL))




class AdvectionOperatorBase(TimeDependentOperator):
    def __init__(self, discr, v, 
            inflow_tag="inflow",
            inflow_u=hedge.data.make_tdep_constant(0),
            outflow_tag="outflow",
            flux_type="central",
            direct_flux=True
            ):
        self.discr = discr
        self.v = v
        self.inflow_tag = inflow_tag
        self.inflow_u = inflow_u
        self.outflow_tag = outflow_tag
        self.flux_type = flux_type

        from hedge.mesh import check_bc_coverage
        check_bc_coverage(discr.mesh, [inflow_tag, outflow_tag])

        self.nabla = discr.nabla
        self.mass = discr.mass_operator
        self.m_inv = discr.inverse_mass_operator
        self.minv_st = discr.minv_stiffness_t

        self.flux = discr.get_flux_operator(self.get_flux(), direct=direct_flux)

    flux_types = [
            "central",
            "upwind",
            "lf"
            ]

    def get_weak_flux(self):
        from hedge.flux import make_normal, FluxScalarPlaceholder, IfPositive

        u = FluxScalarPlaceholder(0)
        normal = make_normal(self.discr.dimensions)

        if self.flux_type == "central":
            return u.avg*numpy.dot(normal, self.v)
        elif self.flux_type == "lf":
            return u.avg*numpy.dot(normal, self.v) \
                    + 0.5*la.norm(self.v)*(u.int - u.ext)
        elif self.flux_type == "upwind":
            return (numpy.dot(normal, self.v)*
                    IfPositive(numpy.dot(normal, self.v),
                        u.int, # outflow
                        u.ext, # inflow
                        ))
        else:
            raise ValueError, "invalid flux type"

    def max_eigenvalue(self):
        return la.norm(self.v)




class StrongAdvectionOperator(AdvectionOperatorBase):
    def get_flux(self):
        from hedge.flux import make_normal, FluxScalarPlaceholder

        u = FluxScalarPlaceholder(0)
        normal = make_normal(self.discr.dimensions)

        return u.int * numpy.dot(normal, self.v) - self.get_weak_flux()

    def rhs(self, t, u):
        from hedge.discretization import pair_with_boundary, cache_diff_results

        bc_in = self.inflow_u.boundary_interpolant(t, self.discr, self.inflow_tag)
        #bc_out = 0.5*self.discr.boundarize_volume_field(u, self.outflow_tag)
        
        return -numpy.dot(self.v, self.nabla*cache_diff_results(u)) + self.m_inv*(
                self.flux * u
                + self.flux * pair_with_boundary(u, bc_in, self.inflow_tag)
                #+ self.flux * pair_with_boundary(u, bc_out, self.outflow_tag)
                )




class WeakAdvectionOperator(AdvectionOperatorBase):
    def get_flux(self):
        return self.get_weak_flux()

    def rhs(self, t, u):
        from hedge.discretization import \
                pair_with_boundary, \
                cache_diff_results

        bc_in = self.inflow_u.boundary_interpolant(t, self.discr, self.inflow_tag)
        bc_out = self.discr.boundarize_volume_field(u, self.outflow_tag)

        return numpy.dot(self.v, self.minv_st*cache_diff_results(u)) - self.m_inv*(
                self.flux*u
                + self.flux * pair_with_boundary(u, bc_in, self.inflow_tag)
                + self.flux * pair_with_boundary(u, bc_out, self.outflow_tag)
                )




class StrongWaveOperator:
    """This operator discretizes the Wave equation S{part}tt u = c^2 S{Delta} u.

    To be precise, we discretize the hyperbolic system

      * S{part}t u - c div v = 0
      * S{part}t v - c grad u = 0

    Note that this is not unique--we could also choose a different sign for M{v}.
    """

    def __init__(self, c, discr, source_f=None, 
            flux_type="upwind",
            dirichlet_tag=hedge.mesh.TAG_ALL,
            neumann_tag=hedge.mesh.TAG_NONE,
            radiation_tag=hedge.mesh.TAG_NONE):
        self.c = c
        self.discr = discr
        self.source_f = source_f

        if self.c > 0:
            self.sign = 1
        else:
            self.sign = -1

        self.dirichlet_tag = dirichlet_tag
        self.neumann_tag = neumann_tag
        self.radiation_tag = radiation_tag

        from hedge.mesh import check_bc_coverage
        check_bc_coverage(discr.mesh, [
            dirichlet_tag,
            neumann_tag,
            radiation_tag])

        from hedge.flux import FluxVectorPlaceholder, make_normal

        dim = discr.dimensions
        w = FluxVectorPlaceholder(1+dim)
        u = w[0]
        v = w[1:]
        normal = make_normal(dim)

        from hedge.tools import join_fields

        flux_weak = join_fields(
                numpy.dot(v.avg, normal),
                u.avg * normal)
        if flux_type == "central":
            pass
        elif flux_type == "upwind":
            # see doc/notes/hedge-notes.tm
            flux_weak -= self.sign*join_fields(
                    0.5*(u.int-u.ext),
                    0.5*(normal * numpy.dot(normal, v.int-v.ext)))
        else:
            raise ValueError, "invalid flux type"

        flux_strong = join_fields(
                numpy.dot(v.int, normal),
                u.int * normal) - flux_weak

        self.flux = discr.get_flux_operator(-self.c*flux_strong)

        self.nabla = discr.nabla
        self.mass = discr.mass_operator
        self.m_inv = discr.inverse_mass_operator

        self.radiation_normals = discr.boundary_normals(self.radiation_tag)

    def rhs(self, t, w):
        from hedge.discretization import pair_with_boundary, cache_diff_results
        from hedge.tools import join_fields, ptwise_dot

        u = w[0]
        v = w[1:]

        dir_bc = join_fields(
                -self.discr.boundarize_volume_field(u, self.dirichlet_tag),
                self.discr.boundarize_volume_field(v, self.dirichlet_tag))

        neu_bc = join_fields(
                self.discr.boundarize_volume_field(u, self.neumann_tag),
                -self.discr.boundarize_volume_field(v, self.neumann_tag))
        
        rad_u = self.discr.boundarize_volume_field(u, self.radiation_tag)
        rad_v = self.discr.boundarize_volume_field(v, self.radiation_tag)
        rad_n = self.radiation_normals
        rad_bc = join_fields(
                0.5*(rad_u - self.sign*ptwise_dot(rad_n, rad_v)),
                0.5*rad_n*(ptwise_dot(rad_n, rad_v) - self.sign*rad_u)
                )

        from hedge.tools import to_obj_array

        rhs = (-join_fields(
            -self.c*numpy.dot(self.nabla, cache_diff_results(v)), 
            -self.c*(self.nabla*cache_diff_results(u))
            ) + to_obj_array(self.m_inv * (
                self.flux*w 
                + self.flux * pair_with_boundary(w, dir_bc, self.dirichlet_tag)
                + self.flux * pair_with_boundary(w, neu_bc, self.neumann_tag)
                + self.flux * pair_with_boundary(w, rad_bc, self.radiation_tag)
                )))

        if self.source_f is not None:
            rhs[0] += self.source_f(t)

        return rhs

    def max_eigenvalue(self):
        return abs(self.c)




class MaxwellOperator(TimeDependentOperator):
    """A 3D Maxwell operator with PEC boundaries.

    Field order is [Ex Ey Ez Hx Hy Hz].
    """

    def __init__(self, discr, epsilon, mu, upwind_alpha=1, 
            pec_tag=hedge.mesh.TAG_ALL, 
            direct_flux=True, current=None):
        from hedge.flux import make_normal, FluxVectorPlaceholder
        from hedge.mesh import check_bc_coverage
        from hedge.discretization import pair_with_boundary
        from math import sqrt
        from hedge.tools import SubsettableCrossProduct, join_fields

        e_subset = self.get_eh_subset()[0:3]
        h_subset = self.get_eh_subset()[3:6]

        e_cross = self.e_cross = SubsettableCrossProduct(
                op2_subset=e_subset, result_subset=h_subset)
        h_cross = self.h_cross = SubsettableCrossProduct(
                op2_subset=h_subset, result_subset=e_subset)

        self.discr = discr

        self.epsilon = epsilon
        self.mu = mu
        self.c = 1/sqrt(mu*epsilon)

        self.pec_tag = pec_tag

        self.current = current

        check_bc_coverage(discr.mesh, [pec_tag])

        dim = discr.dimensions
        normal = make_normal(dim)

        w = FluxVectorPlaceholder(self.count_subset(self.get_eh_subset()))
        e, h = self.split_eh(w)

        Z = sqrt(mu/epsilon)
        Y = 1/Z

        # see Hesthaven/Warburton, Nodal DG, Sec. 6.5
        self.flux = join_fields(
                # flux e, 
                1/epsilon*(
                    -1/2*h_cross(normal, 
                        h.int-h.ext
                        -upwind_alpha/Z*e_cross(normal, e.int-e.ext))
                    ),
                # flux h
                1/mu*(
                    1/2*e_cross(normal, 
                        e.int-e.ext
                        +upwind_alpha/(Y)*h_cross(normal, h.int-h.ext))
                    ),
                )

        self.flux_op = discr.get_flux_operator(self.flux, direct=direct_flux)

        self.nabla = discr.nabla
        self.m_inv = discr.inverse_mass_operator

    def local_op(self, e, h):
        # in conservation form: u_t + A u_x = 0
        def e_curl(field):
            return self.e_cross(self.nabla, field)

        def h_curl(field):
            return self.h_cross(self.nabla, field)

        from hedge.tools import join_fields

        return join_fields(
                - 1/self.epsilon * h_curl(h),
                1/self.mu * e_curl(e),
                )

    def rhs(self, t, w):
        from hedge.tools import cross
        from hedge.discretization import pair_with_boundary, cache_diff_results
        from hedge.tools import join_fields

        e, h = self.split_eh(w)

        pec_bc = join_fields(
                -self.discr.boundarize_volume_field(e, self.pec_tag),
                self.discr.boundarize_volume_field(h, self.pec_tag)
                )

        if self.current is not None:
            j = self.current.volume_interpolant(t, self.discr)
            e_idx = 0 
            for j_idx, use_component in enumerate(self.get_eh_subset()[0:3]):
                if use_component:
                    local_op_fields[e_idx] -= j[j_idx]
                    e_idx += 1
            
        from hedge.tools import to_obj_array
        return - self.local_op(cache_diff_results(e), cache_diff_results(h)) \
                + to_obj_array(self.m_inv*(
                    self.flux_op * w
                    +self.flux_op * pair_with_boundary(w, pec_bc, self.pec_tag)
                    ))

    def assemble_fields(self, e=None, h=None):
        e_components = self.count_subset(self.get_eh_subset()[0:3])
        h_components = self.count_subset(self.get_eh_subset()[3:6])

        if e is None:
            e = [self.discr.volume_zeros() for i in xrange(e_components)]
        if h is None:
            h = [self.discr.volume_zeros() for i in xrange(h_components)]

        from hedge.tools import join_fields
        return join_fields(e, h)

    def split_eh(self, w):
        e_subset = self.get_eh_subset()[0:3]
        h_subset = self.get_eh_subset()[3:6]

        idx = 0

        e = []
        for use_component in e_subset:
            if use_component:
                e.append(w[idx])
                idx += 1

        h = []
        for use_component in h_subset:
            if use_component:
                h.append(w[idx])
                idx += 1

        from hedge.flux import FluxVectorPlaceholder
        from hedge.tools import join_fields

        if isinstance(w, FluxVectorPlaceholder):
            return FluxVectorPlaceholder(scalars=e), FluxVectorPlaceholder(scalars=h)
        elif isinstance(w, numpy.ndarray):
            return join_fields(*e), join_fields(*h)
        else:
            return e, h

    @staticmethod
    def count_subset(subset):
        from pytools import len_iterable
        return len_iterable(uc for uc in subset if uc)

    def get_eh_subset(self):
        """Return a 6-tuple of C{bool}s indicating whether field components 
        are to be computed. The fields are numbered in the order specified
        in the class documentation.
        """
        return 6*(True,)

    def max_eigenvalue(self):
        """Return the largest eigenvalue of Maxwell's equations as a hyperbolic system."""
        from math import sqrt
        return 1/sqrt(self.mu*self.epsilon)




class TMMaxwellOperator(MaxwellOperator):
    """A 2D TM Maxwell operator with PEC boundaries.

    Field order is [Ez Hx Hy].
    """

    def get_eh_subset(self):
        return (
                (False,False,True) # only ez
                +
                (True,True,False) # hx and hy
                )




class TEMaxwellOperator(MaxwellOperator):
    """A 2D TE Maxwell operator with PEC boundaries.

    Field order is [Ex Ey Hz].
    """

    def get_eh_subset(self):
        return (
                (True,True,False) # ex and ey
                +
                (False,False,True) # only hz
                )




class WeakPoissonOperator(Operator, hedge.tools.OperatorBase):
    """Implements the Local Discontinuous Galerkin (LDG) Method for elliptic
    operators.

    See P. Castillo et al., 
    Local discontinuous Galerkin methods for elliptic problems", 
    Communications in Numerical Methods in Engineering 18, no. 1 (2002): 69-75.
    """
    def __init__(self, discr, diffusion_tensor=None, 
            dirichlet_bc=hedge.data.ConstantGivenFunction(), dirichlet_tag="dirichlet",
            neumann_bc=hedge.data.ConstantGivenFunction(), neumann_tag="neumann",
            flux="ip"):
        """Initialize the weak Poisson operator.

        @arg flux: Either C{"ip"} or C{"ldg"} to indicate which type of flux is 
        to be used. IP tends to be faster, and is therefore the default.
        """
        hedge.tools.OperatorBase.__init__(self)

        self.discr = discr

        fs = self.get_weak_flux_set(flux)

        self.flux_u = discr.get_flux_operator(fs.flux_u)
        self.flux_v = discr.get_flux_operator(fs.flux_v)
        self.flux_u_dbdry = discr.get_flux_operator(fs.flux_u_dbdry)
        self.flux_v_dbdry = discr.get_flux_operator(fs.flux_v_dbdry)
        self.flux_u_nbdry = discr.get_flux_operator(fs.flux_u_nbdry)
        self.flux_v_nbdry = discr.get_flux_operator(fs.flux_v_nbdry)

        self.stiff_t = discr.stiffness_t_operator
        self.mass = discr.mass_operator
        self.m_inv = discr.inverse_mass_operator

        from math import sqrt
        from hedge.mesh import check_bc_coverage

        check_bc_coverage(discr.mesh, [dirichlet_tag, neumann_tag])

        # treat diffusion tensor
        if diffusion_tensor is None:
            diffusion_tensor = hedge.data.ConstantGivenFunction(
                    numpy.eye(discr.dimensions))

        if isinstance(diffusion_tensor, hedge.data.ConstantGivenFunction):
            self.diffusion = self.neu_diff = diffusion_tensor.value
        else:
            self.diffusion = diffusion_tensor.volume_interpolant(discr)
            self.neu_diff = diffusion_tensor.boundary_interpolant(discr, neumann_tag)

        self.dirichlet_bc = dirichlet_bc
        self.dirichlet_tag = dirichlet_tag
        self.neumann_bc = neumann_bc
        self.neumann_tag = neumann_tag

        self.neumann_normals = discr.boundary_normals(self.neumann_tag)

    # pylinear operator infrastructure ----------------------------------------
    def size1(self):
        return len(self.discr)

    def size2(self):
        return len(self.discr)

    def apply(self, before, after):
        after[:] = self.op(before)

    # fluxes ------------------------------------------------------------------
    def get_weak_flux_set(self, flux):
        class FluxSet: pass
        fs = FluxSet()

        if flux == "ldg":
            ldg_terms = True
        elif flux == "ip":
            ldg_terms = False
        else:
            raise "Invalid flux type '%s'" % flux

        from hedge.flux import \
                FluxVectorPlaceholder, FluxScalarPlaceholder, \
                make_normal, PenaltyTerm
        from numpy import dot

        dim = self.discr.dimensions
        vec = FluxVectorPlaceholder(1+dim)
        fs.u = u = vec[0]
        fs.v = v = vec[1:]
        normal = make_normal(dim)

        # central flux
        fs.flux_u = u.avg*normal
        fs.flux_v = dot(v.avg, normal)

        if ldg_terms:
            # ldg terms
            ldg_beta = numpy.array([1]*dim)

            fs.flux_u = fs.flux_u - (u.int-u.ext)*0.5*ldg_beta
            fs.flux_v = fs.flux_v + dot((v.int-v.ext)*0.5, ldg_beta)

        # penalty term
        stab_term = PenaltyTerm() * (u.int - u.ext)
        fs.flux_v -= stab_term

        # boundary fluxes
        fs.flux_u_dbdry = normal * u.ext
        fs.flux_v_dbdry = dot(v.int, normal) - stab_term

        fs.flux_u_nbdry = normal * u.int
        fs.flux_v_nbdry = dot(normal, v.ext)

        return fs

    # operator application, rhs prep ------------------------------------------
    def grad(self, u):
        from hedge.discretization import pair_with_boundary, cache_diff_results

        return self.m_inv * (
                - (self.stiff_t * cache_diff_results(u))
                + self.flux_u*u
                + self.flux_u_dbdry*pair_with_boundary(u, 0, self.dirichlet_tag)
                + self.flux_u_nbdry*pair_with_boundary(u, 0, self.neumann_tag)
                )

    def div(self, v, u=None, apply_minv=True):
        """Compute the divergence of v using an LDG operator.

        The divergence computation is unaffected by the scaling
        effected by the diffusion tensor.

        @param apply_minv: Bool specifying whether to compute a complete 
          divergence operator. If False, the final application of the inverse
          mass operator is skipped. This is used in L{op}() in order to reduce
          the scheme M{M^{-1} S u = f} to M{S u = M f}, so that the mass operator
          only needs to be applied once, when preparing the right hand side
          in @L{prepare_rhs}.
        """
        from hedge.discretization import pair_with_boundary, cache_diff_results
        from hedge.tools import join_fields

        dim = self.discr.dimensions

        if u is None:
            u = self.discr.volume_zeros()
        w = join_fields(u, v)

        dirichlet_bc_w = join_fields(0, [0]*dim)
        neumann_bc_w = join_fields(0, [0]*dim)

        result = (
                -numpy.dot(self.stiff_t, cache_diff_results(v))
                + self.flux_v * w
                + self.flux_v_dbdry * pair_with_boundary(w, dirichlet_bc_w, self.dirichlet_tag)
                + self.flux_v_nbdry * pair_with_boundary(w, neumann_bc_w, self.neumann_tag)
                )
        if apply_minv:
            return self.m_inv * result
        else:
            return result

    def op(self, u):
        from hedge.tools import ptwise_dot
        return self.div(
                ptwise_dot(self.diffusion, self.grad(u)), 
                u, apply_minv=False)

    def prepare_rhs(self, rhs):
        """Perform the rhs(*) function in the class description, i.e.
        return a right hand side for the linear system op(u)=rhs(f).
        
        In matrix form, LDG looks like this:
        
        Mv = Cu + g
        Mf = Av + Bu + h

        where v is the auxiliary vector, u is the argument of the operator, f
        is the result of the operator and g and h are inhom boundary data, and
        A,B,C are some operator+lifting matrices

        M f = A Minv(Cu + g) + Bu + h

        so the linear system looks like

        M f = A Minv Cu + A Minv g + Bu + h
        M f - A Minv g - h = (A Minv C + B)u

        So the right hand side we're putting together here is really

        M f - A Minv g - h
        """

        from hedge.discretization import pair_with_boundary, cache_diff_results
        from hedge.tools import join_fields

        dim = self.discr.dimensions

        dtag = self.dirichlet_tag
        ntag = self.neumann_tag

        dirichlet_bc_u = self.dirichlet_bc.boundary_interpolant(self.discr, dtag)
        vpart = self.m_inv * (
                (self.flux_u_dbdry*pair_with_boundary(0, dirichlet_bc_u, dtag))
                )

        from hedge.tools import ptwise_dot
        diff_v = ptwise_dot(self.diffusion, vpart)

        def neumann_bc_v():
            return ptwise_dot(self.neu_diff, 
                    self.neumann_normals*
                        self.neumann_bc.boundary_interpolant(self.discr, ntag))

        w = join_fields(0, diff_v)
        dirichlet_bc_w = join_fields(dirichlet_bc_u, [0]*dim)
        neumann_bc_w = join_fields(0, neumann_bc_v())

        return self.discr.mass_operator * rhs.volume_interpolant(self.discr) - (
                -numpy.dot(self.stiff_t, cache_diff_results(diff_v))
                + self.flux_v * w
                + self.flux_v_dbdry * pair_with_boundary(w, dirichlet_bc_w, dtag)
                + self.flux_v_nbdry * pair_with_boundary(w, neumann_bc_w, ntag)
                )

    def grad_matrix(self):
        discr = self.discr
        dim = discr.dimensions

        def assemble_local_vstack(operators):
            n = len(operators)
            dof = len(discr)
            result = pyublas.zeros((n*dof, dof), flavor=pyublas.SparseBuildMatrix)

            from hedge._internal import MatrixTarget
            tgt = MatrixTarget(result, 0, 0)

            for i, op in enumerate(operators):
                op.perform_on(tgt.rebased_target(i*dof, 0))
            return result

        def assemble_local_hstack(operators):
            n = len(operators)
            dof = len(discr)
            result = pyublas.zeros((dof, n*dof), flavor=pyublas.SparseBuildMatrix)

            from hedge._internal import MatrixTarget
            tgt = MatrixTarget(result, 0, 0)

            for i, op in enumerate(operators):
                op.perform_on(tgt.rebased_target(0, i*dof))
            return result

        def assemble_local_diag(operators):
            n = len(operators)
            dof = len(discr)
            result = pyublas.zeros((n*dof, n*dof), flavor=pyublas.SparseBuildMatrix)

            from hedge._internal import MatrixTarget
            tgt = MatrixTarget(result, 0, 0)

            for i, op in enumerate(operators):
                op.perform_on(tgt.rebased_target(i*dof, i*dof))
            return result

        def fast_mat(mat):
            return pyublas.asarray(mat, flavor=pyublas.SparseExecuteMatrix)

        def assemble_grad():
            n = self.discr.dimensions
            dof = len(discr)

            minv = fast_mat(assemble_local_diag([self.m_inv] * dim))

            m_local_grad = fast_mat(-assemble_local_vstack(self.discr.minv_stiffness_t))

            fluxes = pyublas.zeros((n*dof, dof), flavor=pyublas.SparseBuildMatrix)
            from hedge._internal import MatrixTarget
            fluxes_tgt = MatrixTarget(fluxes, 0, 0)
            self.flux_u.perform_inner(fluxes_tgt)
            self.flux_u_dbdry.perform_int_bdry(self.dirichlet_tag, fluxes_tgt)
            self.flux_u_nbdry.perform_int_bdry(self.neumann_tag, fluxes_tgt)

            return m_local_grad + minv * fast_mat(fluxes)

        return assemble_grad()





class StrongHeatOperator(TimeDependentOperator):
    def __init__(self, discr, coeff=hedge.data.ConstantGivenFunction(1), 
            dirichlet_bc=hedge.data.ConstantGivenFunction(), dirichlet_tag="dirichlet",
            neumann_bc=hedge.data.ConstantGivenFunction(), neumann_tag="neumann",
            ldg=True):
        self.discr = discr

        fs = self.get_strong_flux_set(ldg)

        self.flux_u = discr.get_flux_operator(fs.flux_u)
        self.flux_v = discr.get_flux_operator(fs.flux_v)
        self.flux_u_dbdry = discr.get_flux_operator(fs.flux_u_dbdry)
        self.flux_v_dbdry = discr.get_flux_operator(fs.flux_v_dbdry)
        self.flux_u_nbdry = discr.get_flux_operator(fs.flux_u_nbdry)
        self.flux_v_nbdry = discr.get_flux_operator(fs.flux_v_nbdry)

        self.nabla = discr.nabla
        self.stiff = discr.stiffness_operator
        self.mass = discr.mass_operator
        self.m_inv = discr.inverse_mass_operator

        from hedge.mesh import check_bc_coverage
        check_bc_coverage(discr.mesh, [dirichlet_tag, neumann_tag])

        def fast_diagonal_mat(vec):
            return num.diagonal_matrix(vec, flavor=num.SparseExecuteMatrix)

        self.sqrt_coeff = numpy.sqrt(
                coeff.volume_interpolant(discr))
        self.dir_sqrt_coeff = numpy.sqrt(
                coeff.boundary_interpolant(discr, dirichlet_tag))
        self.neu_sqrt_coeff = numpy.sqrt(
                coeff.boundary_interpolant(discr, neumann_tag))

        self.dirichlet_bc = dirichlet_bc
        self.dirichlet_tag = dirichlet_tag
        self.neumann_bc = neumann_bc
        self.neumann_tag = neumann_tag

        self.neumann_normals = discr.boundary_normals(self.neumann_tag)

    # fluxes ------------------------------------------------------------------
    def get_weak_flux_set(self, ldg):
        class FluxSet: pass
        fs = FluxSet()

        from hedge.flux import FluxVectorPlaceholder, FluxScalarPlaceholder, make_normal

        # note here:

        # local DG is unlike the other kids in that the computation of the flux
        # of u depends *only* on u, whereas the computation of the flux of v
        # (yielding the final right hand side) may also depend on u. That's why
        # we use the layout [u,v], where v is simply omitted for the u flux
        # computation.

        dim = self.discr.dimensions
        vec = FluxVectorPlaceholder(1+dim)
        fs.u = u = vec[0]
        fs.v = v = vec[1:]
        normal = fs.normal = make_normal(dim)

        # central
        fs.flux_u = u.avg*normal
        fs.flux_v = numpy.dot(v.avg, normal)

        # dbdry is "dirichlet boundary"
        # nbdry is "neumann boundary"
        fs.flux_u_dbdry = fs.flux_u
        fs.flux_u_nbdry = fs.flux_u

        fs.flux_v_dbdry = fs.flux_v
        fs.flux_v_nbdry = fs.flux_v

        if ldg:
            ldg_beta = numpy.ones((dim,))

            fs.flux_u = fs.flux_u - (u.int-u.ext)*0.5*ldg_beta
            fs.flux_v = fs.flux_v + numpy.dot((v.int-v.ext)*0.5, ldg_beta)

        return fs

    def get_strong_flux_set(self, ldg):
        fs = self.get_weak_flux_set(ldg)

        u = fs.u
        v = fs.v
        normal = fs.normal

        fs.flux_u = u.int*normal - fs.flux_u
        fs.flux_v = numpy.dot(v.int, normal) - fs.flux_v
        fs.flux_u_dbdry = u.int*normal - fs.flux_u_dbdry
        fs.flux_v_dbdry = numpy.dot(v.int, normal) - fs.flux_v_dbdry
        fs.flux_u_nbdry = u.int*normal - fs.flux_u_nbdry
        fs.flux_v_nbdry = numpy.dot(v.int, normal) - fs.flux_v_nbdry

        return fs

    # boundary conditions -----------------------------------------------------
    def dirichlet_bc_u(self, t, sqrt_coeff_u):
        return (
                -self.discr.boundarize_volume_field(sqrt_coeff_u, self.dirichlet_tag)
                +2*self.dir_sqrt_coeff*self.dirichlet_bc.boundary_interpolant(
                    t, self.discr, self.dirichlet_tag)
                )

    def dirichlet_bc_v(self, t, sqrt_coeff_v):
        return self.discr.boundarize_volume_field(sqrt_coeff_v, self.dirichlet_tag)

    def neumann_bc_u(self, t, sqrt_coeff_u):
        return self.discr.boundarize_volume_field(sqrt_coeff_u, self.neumann_tag)

    def neumann_bc_v(self, t, sqrt_coeff_v):
        return (
                -self.discr.boundarize_volume_field(sqrt_coeff_v, self.neumann_tag)
                +
                2*self.neumann_normals*
                self.neumann_bc.boundary_interpolant(t, self.discr, self.neumann_tag)
                )

    # right-hand side ---------------------------------------------------------
    def rhs(self, t, u):
        from hedge.discretization import pair_with_boundary, cache_diff_results
        from math import sqrt
        from hedge.tools import join_fields

        dtag = self.dirichlet_tag
        ntag = self.neumann_tag

        sqrt_coeff_u = self.sqrt_coeff * u

        dirichlet_bc_u = self.dirichlet_bc_u(t, sqrt_coeff_u)
        neumann_bc_u = self.neumann_bc_u(t, sqrt_coeff_u)

        v = self.m_inv * (
                self.sqrt_coeff*(self.stiff * cache_diff_results(u))
                - self.flux_u*sqrt_coeff_u
                - self.flux_u_dbdry*pair_with_boundary(sqrt_coeff_u, dirichlet_bc_u, dtag)
                - self.flux_u_nbdry*pair_with_boundary(sqrt_coeff_u, neumann_bc_u, ntag)
                )
        sqrt_coeff_v = self.sqrt_coeff * v

        dirichlet_bc_v = self.dirichlet_bc_v(t, sqrt_coeff_v)
        neumann_bc_v = self.neumann_bc_v(t, sqrt_coeff_v)

        w = join_fields(sqrt_coeff_u, sqrt_coeff_v)
        dirichlet_bc_w = join_fields(dirichlet_bc_u, dirichlet_bc_v)
        neumann_bc_w = join_fields(neumann_bc_u, neumann_bc_v)

        return self.m_inv * (
                numpy.dot(self.stiff, cache_diff_results(self.sqrt_coeff*v))
                - self.flux_v * w
                - self.flux_v_dbdry * pair_with_boundary(w, dirichlet_bc_w, dtag)
                - self.flux_v_nbdry * pair_with_boundary(w, neumann_bc_w, ntag)
                )



# Hedge - the Hybrid'n'Easy DG Environment
# Copyright (C) 2007 Andreas Kloeckner
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.




from __future__ import division
import numpy
import numpy.linalg as la




def main() :
    from hedge.element import \
            IntervalElement, \
            TriangularElement, \
            TetrahedralElement
    from hedge.timestep import RK4TimeStepper, AdamsBashforthTimeStepper
    from hedge.visualization import SiloVisualizer, VtkVisualizer
    from pytools.stopwatch import Job
    from math import sin, cos, pi, exp, sqrt
    from hedge.parallel import guess_parallelization_context

    pcon = guess_parallelization_context()

    dim = 2

    if dim == 1:
        if pcon.is_head_rank:
            from hedge.mesh import make_uniform_1d_mesh
            mesh = make_uniform_1d_mesh(-10, 10, 500)

        el_class = IntervalElement
    elif dim == 2:
        from hedge.mesh import \
                make_disk_mesh, \
                make_regular_square_mesh, \
                make_square_mesh, \
                make_rect_mesh
        if pcon.is_head_rank:
            #mesh = make_disk_mesh(max_area=5e-3)
            #mesh = make_regular_square_mesh(
                    #n=9, periodicity=(True,True))
            mesh = make_rect_mesh(a=(-0.5,-0.5),b=(3.5,0.5),max_area=0.008)
            #mesh.transform(Rotation(pi/8))
        el_class = TriangularElement
    elif dim == 3:
        if pcon.is_head_rank:
            mesh = make_ball_mesh(max_volume=0.0005)
        el_class = TetrahedralElement
    else:
        raise RuntimeError, "bad number of dimensions"

    if pcon.is_head_rank:
        print "%d elements" % len(mesh.elements)
        mesh_data = pcon.distribute_mesh(mesh)
    else:
        mesh_data = pcon.receive_mesh()

    discr = pcon.make_discretization(mesh_data, el_class(7))
    stepper = RK4TimeStepper()
    #stepper = AdamsBashforthTimeStepper(1)
    #vis = VtkVisualizer(discr, pcon, "fld")
    vis = SiloVisualizer(discr, pcon)

    def source_u(x):
        return exp(-numpy.dot(x, x)*128)

    source_u_vec = discr.interpolate_volume_function(source_u)

    def source_vec_getter(t):
        from math import sin
        return source_u_vec*sin(10*t)

    from hedge.operators import StrongWaveOperator
    from hedge.mesh import TAG_ALL, TAG_NONE
    op = StrongWaveOperator(-1, discr, 
            source_vec_getter,
            dirichlet_tag=TAG_ALL,
            neumann_tag=TAG_NONE,
            radiation_tag=TAG_NONE,
            flux_type="upwind",
            )

    from hedge.tools import join_fields
    fields = join_fields(discr.volume_zeros(),
            [discr.volume_zeros() for i in range(discr.dimensions)])
    #fields = join_fields(
            #discr.interpolate_volume_function(lambda x: sin(x[0])),
            #[discr.volume_zeros() for i in range(discr.dimensions)]) # v

    dt = discr.dt_factor(op.max_eigenvalue())
    nsteps = int(10/dt)
    if pcon.is_head_rank:
        print "dt", dt
        print "nsteps", nsteps

    # diagnostics setup -------------------------------------------------------
    from pytools.log import LogManager, \
            add_general_quantities, \
            add_simulation_quantities, \
            add_run_info

    logmgr = LogManager("wave.dat", "w", pcon.communicator)
    add_run_info(logmgr)
    add_general_quantities(logmgr)
    add_simulation_quantities(logmgr, dt)
    discr.add_instrumentation(logmgr)

    from pytools.log import IntervalTimer
    vis_timer = IntervalTimer("t_vis", "Time spent visualizing")
    logmgr.add_quantity(vis_timer)
    stepper.add_instrumentation(logmgr)

    from hedge.log import Integral, LpNorm
    u_getter = lambda: fields[0]
    logmgr.add_quantity(LpNorm(u_getter, discr, 1, name="l1_u"))
    logmgr.add_quantity(LpNorm(u_getter, discr, name="l2_u"))

    logmgr.add_watches(["step.max", "t_sim.max", "l2_u", "t_step.max"])

    # timestep loop -----------------------------------------------------------
    for step in range(nsteps):
        logmgr.tick()

        t = step*dt

        if step % 10 == 0:
            visf = vis.make_file("fld-%04d" % step)

            vis.add_data(visf,
                    [
                        ("u", fields[0]),
                        ("v", fields[1:]), 
                        ("n", discr.volumize_boundary_field(
                            discr.boundary_normals())),
                    ],
                    time=t,
                    #scale_factor=2e1,
                    step=step)
            visf.close()

        def rhs(t, y):
            return op.rhs(t, y) - 3*join_fields(fields[0], 0*fields[1:])

        fields = stepper(fields, t, dt, rhs)

    vis.close()

    logmgr.tick()
    logmgr.save()

if __name__ == "__main__":
    #import cProfile as profile
    #profile.run("main()", "wave2d.prof")
    main()

